from .main import *
from .para import *
from .tensor_scaling import rescale_lambda_centerized_workhorse
from .tensor_scaling import rescale_lambda_centerized

from .hdf5_io import *
from .coil import *
from .exp import *
from .util import *
from .freesurfer import *
from .roi import *
from .subject import *
from .postproc import *
from .simnibs import *
import pynibs.models
from .exp import *
from .opt import *
from .neuron import *

try:
    from .pckg import libeep
except (ImportError, SyntaxError):
    # print("libeep not found. visor import will not work.")
    pass
try:
    from paraview.simple import *
except ImportError:
    # print("Paraview not installed. Plot functionalities will not work.")
    pass


""" Functions to import data from ANT Visor 2 / ANT EEG software go here """
import warnings

import os
import h5py
import numpy as np
from pynibs import write_arr_to_hdf5
from pynibs.util.util import quat_to_rot
from scipy.spatial.transform import Rotation as R
from scipy import signal
import pandas as pd

try:
    from pynibs.pckg import libeep
except (ImportError, SyntaxError):
    pass
    # print("lipbeep not found. visor import will not work.")


def read_nlr(fname):
    """
    Reads NLR coordinates from *_recording.mri file

    Parameters
    ----------
    fname

    Returns
    -------
    fiducials : np.array of float [3 x 3]
        The rows contain the fiducial points in ANT nifit space (nasion, left ear, right ear)
    """
    f = open(fname, "r")
    text = f.readlines()

    fiducials = np.empty((3, 3))

    for i, line in enumerate(text):
        # nasion
        if "VoxelOnPositiveXAxis" in line:
            line = text[i + 1].replace("\t", " ")
            line = line.replace("\n", "")
            fiducials[0, :] = np.array([int(t) for t in line.split(" ")])

        # left ear
        if "VoxelOnNegativeYAxis" in line:
            line = text[i + 1].replace("\t", " ")
            line = line.replace("\n", "")
            fiducials[1, :] = np.array([int(t) for t in line.split(" ")])

        # right ear
        if "VoxelOnPositiveYAxis" in line:
            line = text[i + 1].replace("\t", " ")
            line = line.replace("\n", "")
            fiducials[2, :] = np.array([int(t) for t in line.split(" ")])

    return fiducials


def get_instrument_marker(im_path, verbose=False):
    """
    Return all instrument markers from visor .cnt file
    Coordinate system in raw ANT space (NLR) defined as:

    origin: intersection between line of ear fiducials and nasion
    x-axis: origin -> nasion
    y-axis: origin -> left ear
    z-axis: origin -> superior

    Parameters
    ----------
    im_path : str
        Path to instrument-marker-file .cnt file
    verbose: bool
        Some verbosity messages (default: False)

    Returns
    -------
    im_list : list of dict
        List containing stimulation parameters:
        - coil_mean_raw [4 x 4 matrix]
        - StimulusID
        - etc...

    """
    f = libeep.read_cnt(im_path)
    n_trig = f.get_trigger_count()
    # some triggers (2?) are some other information
    # so only take the ones with 'StimulusID' at 3rd position
    ims = [f.get_trigger(i)[3] for i in range(n_trig) if "StimulusID" in f.get_trigger(i)[3]]
    # or: if f.get_trigger(i)[0] == '6'

    if verbose:
        print(f"Found {len(ims)} instrument markers.")
    assert len(ims), "No instrument markers found in file"

    # now build list of matsimnibs from the instrument markers
    data = []

    for i, im in enumerate(ims):

        # transform string from .cnt file to dictionary
        data.append(dict(item.split('=') for item in im.split()[1:] if '=' in item))

        # floatify numeric variables
        for key in data[-1].keys():
            try:
                if key == "StimulusID":
                    data[-1][key] = int(data[-1][key])
                else:
                    data[-1][key] = float(data[-1][key])
            except ValueError:
                pass

        # transform to SimNIBS raw format
        matsimnibs_raw = np.zeros((4, 4))
        matsimnibs_raw[3, 3] = 1
        matsimnibs_raw[0:3, 3] = np.array([data[-1]['PosX'], data[-1]['PosY'], data[-1]['PosZ']]) * 1000
        quat = np.array([data[-1]['QuatX'], data[-1]['QuatY'], data[-1]['QuatZ'], data[-1]['QuatW']])
        matsimnibs_raw[0:3, 0:3] = R.from_quat(quat).as_dcm()
        data[-1]["coil_mean_raw"] = matsimnibs_raw

    return data


def get_cnt_data(fn, channels='all', trigger_val='1', max_duration=10,
                 fn_hdf5=None, path_hdf5=None, verbose=False, return_data=False):
    """
    Reads ANT .cnt EMG/EEG data file and chunks timesieries into triggerN - trigggerN+1
    Can directly write the zaps into hdf5 if argument is provided.
    Starting with the first trigger and ending with get_sample_count()-1

    Parameters
    ----------
    fn: str
        .cnt Filename
    channels: str | int | list of int | list of str
        Which channel(s) to return. Default: all.
        Can be channel number(s) or channel name(s).
    trigger_val: str
        Trigger value to read as zap trigger. Default: '1'
    max_duration : int
        Maximum duration in [s] per chunk. Rest is dropped.
    fn_hdf5: str, optional, default: None
        If given, cnt data is written into hdf5 file under "path_hdf5" as pandas dataframe with column name "qoi_name"
        and nothing is returned. Default: None
    path_hdf5: str, optional, default: None
        If fn_hdf5, path of pandas dataframe where the data is saved (e.g. "/phys_data/raw/EEG")
    verbose: bool
        Some verbosity messages (default: False)
    return_data: bool
        Return data as list of np.array

    Returns
    -------
    data_lst: list of np.ndarray with shape=(samples,channels), optional
        List of EEG/EMG data. Only returned if fn_hdf5 is not None
    """

    f = libeep.read_cnt(fn)
    n_trig = f.get_trigger_count()
    n_samples = f.get_sample_count()
    n_channels = f.get_channel_count()
    sf = f.get_sample_frequency()
    chan_names = [f.get_channel(i)[0].lower() for i in range(n_channels)]

    if channels == 'all' or isinstance(channels, list) and channels[0] == 'all':
        channels_idx = range(n_channels)
    elif isinstance(channels, int):
        channels_idx = [channels]
    elif isinstance(channels, str):
        channels_idx = chan_names.index(channels.lower())
    elif type(channels) == list and all(type(chan) == int for chan in channels):
        channels_idx = channels
        assert np.all(np.array(channels_idx) >= 0), "Only positive channels numbers allowd"
        assert np.max(np.array(channels_idx)) < n_channels, f"Only {n_channels} channels found."

    elif type(channels) == list and all(type(chan) == str for chan in channels):
        channels_idx = [chan_names.index(chan.lower()) for chan in channels]

    else:
        raise NotImplementedError("Channels must be 'all', list(int), list(str)")

    assert channels_idx, "No channels with name / idx found."

    if fn_hdf5 is not None:
        assert path_hdf5, "Please provide path_hdf5="

    if verbose:
        print(f"get_cnt_data: {n_trig} triggers found.")
        print(f"get_cnt_data: {n_samples} samples found.")
        print(f"get_cnt_data: {sf} Hz sampling frequency.")
        print(f"get_cnt_data: {n_channels} channels found.")

    # get data between samples
    data_lst = []

    # chunk into triggers
    trigger_idx = 0
    arr_idx = 0
    last_zap_done = False
    trigger_zap = 0
    # we want the data between trigger and trigger+1
    while trigger_idx < n_trig - 1:

        try:
            start = f.get_trigger(trigger_idx)

            # only use the triggers that have the correct trifger value
            if start[0] != trigger_val:
                if verbose:
                    print(f"get_cnt_data: Skipping idx {trigger_idx}: {start} (start)")
                trigger_idx += 1
                continue
            end = f.get_trigger(trigger_idx + 1)
            # also trigger+1 needs to have the correct trigger_val
            while end[0] != trigger_val:
                if verbose:
                    print(f"Skipping idx {trigger_idx}: {start} (end)")
                trigger_idx += 1
                if trigger_idx >= n_trig - 1:
                    break
                end = f.get_trigger(trigger_idx)

            # some sanity checks
            if not start[1] < end[1]:
                if verbose:
                    print(f"Trigger {trigger_idx} and {trigger_idx + 1}: wrong sample number "
                              f"({trigger_idx}: {start[1]}, {trigger_idx + 1}: {end[1]}]")
                # the eeg cnt files and with a trigger. get data from trigger to end-offile
                if trigger_idx == n_trig - 1:
                    end = (end[0], f.get_sample_count())
                    last_zap_done = True

            assert start[1] < (end[1] - 1), \
                f"Trigger {trigger_idx} and {trigger_idx + 1}: too close together " \
                f"({trigger_idx}: {start[1]}, {trigger_idx + 1}: {end[1]}]"

            # get sample number from trigger-tuple
            start = start[1]
            end = end[1] - 1
            length_org = end - start

            # cut to max duration chunk length
            end = np.min((end, start + sf * max_duration))

            if verbose:
                print(f"get_cnt_data: Trigger {trigger_idx:0>3}: {float(length_org) / sf:2.2}s / "
                      f"{float(end - start) / sf:0.2}s")
            data = f.get_samples(start, end)
            data_res = np.reshape(data, (end - start, n_channels), order='F')

            if return_data:
                data_lst.append(data_res)

        except (SystemError, UnicodeDecodeError) as e:
            print(f"Trigger {trigger_idx} error")
            print(e)
            continue

        if fn_hdf5 is not None:
            with h5py.File(fn_hdf5, "a") as fi:
                fi[path_hdf5 + f"/{trigger_zap:04d}"] = data_res
                trigger_zap += 1

        trigger_idx += 1

    # grap data for the last zap (trigger to end_of_file
    if not last_zap_done:

        try:
            start = f.get_trigger(trigger_idx)

            # only use the triggers that have the correct trigger value
            if start[0] != trigger_val:
                if verbose:
                    print(f"get_cnt_data: Skipping idx {trigger_idx}: {start} (start)")
                trigger_idx += 1
            end = f.get_sample_count()

            assert start[1] < (end - 1), \
                f"Trigger {trigger_idx} and {trigger_idx + 1}: too close together " \
                f"({trigger_idx}: {start[1]}, {trigger_idx + 1}: {end}]"

            # get sample number from trigger-tuple
            start = start[1]
            length_org = end - start

            # cut to max duration chunk length
            end = np.min((end, start + sf * max_duration))

            if verbose:
                print(f"get_cnt_data: Trigger {trigger_idx:0>3}: {float(length_org) / sf:2.2}s / "
                      f"{float(end - start) / sf:0.2}s")
            data = f.get_samples(start, end)
            data_res = np.reshape(data, (end - start, n_channels), order='F')

            if return_data:
                data_lst.append(data_res)

            if fn_hdf5 is not None:
                with h5py.File(fn_hdf5, "a") as fi:
                    fi[path_hdf5 + f"/{trigger_zap:04d}"] = data_res
                    trigger_zap += 1

            trigger_idx += 1

        except (SystemError, UnicodeDecodeError) as e:
            print(f"Trigger {trigger_idx} error")
            print(e)

        # reshape according to channel count
        # [chan1, chan2, chan3, chan1, chan2, chan3]
        # data_res = np.reshape(data, (end - start, n_channels), order='F')
        #
        # if return_data:
        #     data_lst.append(data_res[:, channels_idx])
        #
        # if fn_hdf5 is not None:
        #     with h5py.File(fn_hdf5, "a") as fi:
        #         fi[path_hdf5 + f"/{trigger_idx:04d}"] = data_res[:, channels_idx]

    # append last chunk
    # start = f.get_trigger(n_trig - 2)[1]
    # end = n_samples - 1
    # end = np.min((end, start + sf * max_duration))  # cut to max chunk length
    #
    # data = f.get_samples(start, end)
    # data_res = np.reshape(data, (end - start, n_channels), order='F')
    # if fn_hdf5:
    #     write_arr_to_hdf5(fn_hdf5=fn_hdf5,
    #                       arr_name=arr_name.format(arr_idx),
    #                       data=data_res[:, channels_idx],
    #                       verbose=verbose)
    # else:

    if return_data:
        return data_lst


def filter_emg(emg, fs):
    """
    Filter emg signals

    Parameters
    ----------
    emg : list of np.array [n_stimuli]
        Raw EMG data. Each list entry contains a np.array of size [n_samples x n_channel]
        Each channel is filtered in the same way.
    fs : float
        Sampling frequency

    Returns
    -------
    emg_filt : list of np.array [n_stimuli]
        Filtered EMG data
    """

    # 5 Hz Butterworth high pass
    ############################
    b_butterhigh, a_butterhigh = signal.butter(N=5, Wn=5, btype='high', analog=False, fs=fs)
    # plot_frequency_response(a_butterhigh, b_butterhigh, fs=fs)

    # 200 Hz Butterworth low pass
    ############################
    b_butterlow, a_butterlow = signal.butter(N=5, Wn=200, btype='low', analog=False, fs=fs)
    # plot_frequency_response(a_butterlow, b_butterlow, fs=fs)

    # 50 Hz Notch filter
    ############################
    b_notch50, a_notch50 = signal.iirnotch(w0=50 / (fs / 2), Q=30)
    # plot_frequency_response(a_notch50, b_notch50, fs=fs)

    # 100 Hz Notch filter
    ############################
    b_notch100, a_notch100 = signal.iirnotch(w0=100 / (fs / 2), Q=50)
    # plot_frequency_response(a_notch100, b_notch100, fs=fs)

    # 150 Hz Notch filter
    ############################
    b_notch150, a_notch150 = signal.iirnotch(w0=150 / (fs / 2), Q=30)
    # plot_frequency_response(a_notch150, b_notch150, fs=fs)

    # 200 Hz Notch filter
    ############################
    b_notch200, a_notch200 = signal.iirnotch(w0=200 / (fs / 2), Q=30)
    # plot_frequency_response(a_notch200, b_notch200, fs=fs)

    # Filter signals
    emg_filt = []

    for e in emg:
        emg_filt.append(np.zeros(e.shape))
        for i_channel in range(e.shape[1]):
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch50, a_notch50, e[:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch100, a_notch100, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch150, a_notch150, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch200, a_notch200, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_butterlow, a_butterlow, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_butterhigh, a_butterhigh, emg_filt[-1][:, i_channel])
            emg_filt[-1][:, i_channel] = signal.filtfilt(b_notch50, a_notch50, emg_filt[-1][:, i_channel])

    return emg_filt



import numpy as np
import re


def read_channel_names(fname):
    """
    Reads the channel names from .vhdr (info) file, which is recorded during EEG

    Parameters
    ----------
    fname : str
        Filename of .vhdr info file

    Returns
    -------
    channel_names : list of str
        List containing the channel names
    """

    f = open(fname, "r")
    eof = False
    i_line = 0
    channel_info_start = np.inf
    channel_names = []

    while not eof:
        l = f.readline()

        if l == "[Channel Infos]\n":
            channel_info_start = 5 + i_line

        if i_line >= channel_info_start:

            if l == "\n":
                break

            channel_names.append(re.search("(?<=\=)(.*?)(?=\,)", l).group(0))

        i_line += 1

    return channel_names


def read_sampling_frequency(fname):
    """
    Reads the sampling frequency from .vhdr (info) file, which is recorded during EEG

    Parameters
    ----------
    fname : str
        Filename of .vhdr info file

    Returns
    -------
    Sampling_frequency : float
        Sampling frequency
    """

    f = open(fname, "r")
    eof = False

    while not eof:
        l = f.readline()

        if "Sampling Rate [Hz]: " in l:
            start_idx = l.find(":") + 2
            sampling_frequency = float(l[start_idx:])
            break

    return sampling_frequency



""" Functions to import data from Localite TMS Navigator """

from xml.etree import ElementTree as xmlt
import numpy as np


def get_tms_elements(xml_paths, verbose=False):
    """
    Read needed elements out of the tms-xml-file.

    Parameters
    ----------
    xml_paths : list of str or str
        Paths to coil0-file and optionally coil1-file if there is no coil1-file, use empty string
    verbose : bool, optional, default: False
        Print output messages

    Returns
    -------
    coils_array : nparray of float [3xNx4x4]
        Coil0, coil1 and mean-value of N 4x4 coil-arrays
    ts_tms_lst : list of int [N]
        Timestamps of valid tms-measurements
    current_lst : list of int [N]
        Measured currents
    idx_invalid : list of int
        List of indices of invalid coil positions (w.r.t. all timestamps incl invalid)
    """
    if isinstance(xml_paths, str):
        xml_paths = [xml_paths]
    # handle case if there is no coil1
    if len(xml_paths) > 1 and not xml_paths[1]:
        xml_paths[1] = xml_paths[0]
    if len(xml_paths) == 1:
        xml_paths.append(xml_paths[0])

    # allocate new array and lists
    coils_array, ts_tms_lst, current_lst = np.empty([3, 0, 4, 4]), [], []

    # parse XML document
    coil0_tree, coil1_tree = xmlt.parse(xml_paths[0]), xmlt.parse(xml_paths[1])
    coil0_root, coil1_root = coil0_tree.getroot(), coil1_tree.getroot()

    # iterate over all 'TriggerMarker' tags
    i_stim = 0
    idx_invalid = []

    for coil0_tm, coil1_tm in zip(coil0_root.iter('TriggerMarker'), coil1_root.iter('TriggerMarker')):
        coil_array = np.empty([0, 1, 4, 4])

        # get tag were the matrix is
        coil0_ma, coil1_ma = coil0_tm.find('Matrix4D'), coil1_tm.find('Matrix4D')

        # get coil0
        coil_array = np.append(coil_array, read_coil(coil0_ma), axis=0)

        # if present, get coil1
        if xml_paths[0] == xml_paths[1]:
            coil_array = np.append(coil_array, np.identity(4)[np.newaxis, np.newaxis, :, :], axis=0)
        else:
            coil_array = np.append(coil_array, read_coil(coil1_ma), axis=0)

        # check for not valid coils and calculate mean value
        if not np.allclose(coil_array[0, 0, :, :], np.identity(4)) and \
                not np.allclose(coil_array[1, 0, :, :], np.identity(4)):
            coil_array = np.append(coil_array,
                                   np.expand_dims((coil_array[0, :, :, :] + coil_array[1, :, :, :]) / 2, axis=0),
                                   axis=0)
        elif np.allclose(coil_array[0, 0, :, :], np.identity(4)) and not np.allclose(coil_array[1, 0, :, :],
                                                                                     np.identity(4)):
            coil_array = np.append(coil_array, np.expand_dims(coil_array[1, :, :, :], axis=0), axis=0)
        elif np.allclose(coil_array[1, 0, :, :], np.identity(4)) and not np.allclose(coil_array[0, 0, :, :],
                                                                                     np.identity(4)):
            coil_array = np.append(coil_array, np.expand_dims(coil_array[0, :, :, :], axis=0), axis=0)
        else:
            idx_invalid.append(i_stim)
            if verbose:
                print("Removing untracked (and possibly accidental) coil position #{} (identity matrix)".format(i_stim))
            i_stim += 1
            continue

        # print(i_stim)
        i_stim += 1

        coils_array = np.append(coils_array, coil_array, axis=1)

        # get timestamp
        ts_tms_lst.append(int(coil0_tm.get('recordingTime')))

        # get current
        xml_rv = coil0_tm.find('ResponseValues')
        xml_va = xml_rv.findall('Value')

        # if valueA is NaN, compute dI/dt with amplitudeA
        if xml_va[0].get('response') == 'NaN':
            current_lst.append(str(round(float(xml_va[2].get('response')) * 1.4461)))  # was 1.38
        else:
            current_lst.append(xml_va[0].get('response'))

    return [coils_array, ts_tms_lst, current_lst, idx_invalid]


def read_coil(xml_ma):
    """
    Read coil-data from xml element.

    Parameters
    ----------
    xml_ma : xml-element
        Coil data

    Returns
    -------
    coil : nparray of float [4x4]
        Coil elements
    """
    # index2 for all coils from triggermarker
    coil = np.empty([1, 1, 4, 4])
    for coil_index1 in range(4):
        for coil_index2 in range(4):
            coil[0, 0, coil_index1, coil_index2] = (float(xml_ma.get('data' + str(coil_index1) + str(coil_index2))))
    return coil


def match_instrument_marker_file(xml_paths, im_path):
    """ Assign right instrument marker condition to every triggermarker (get instrument marker out of file).

    Parameters
    ----------
    xml_paths : list of str
        Paths to coil0-file and optionally coil1-file if there is no coil1-file, use empty string
    im_path : str
        Path to instrument-marker-file

    Returns
    -------
    coil_cond_lst : list of str
        Right conditions
    """

    tm_array, tms_time_arr, tms_cur_arr, tms_idx_invalid = get_tms_elements(xml_paths)
    # get coil mean value
    tm_array = tm_array[2]
    im_array, im_cond_lst = get_instrument_marker(im_path)

    # get indices of conditions
    im_index_lst, drop_idx = match_tm_to_im(tm_array, im_array, tms_time_arr, tms_cur_arr)

    # list to save conditions
    coil_cond_lst = []

    for cond_index in im_index_lst:
        coil_cond_lst.append(im_cond_lst[cond_index])
    return coil_cond_lst, drop_idx


def match_instrument_marker_string(xml_paths, condition_order):
    """
    Assign right instrument marker condition to every triggermarker (get instrument marker out of list of strings).

    Parameters
    ----------
    xml_paths : list of str
        Paths to coil0-file and optionally coil1-file if there is no coil1-file, use empty string
    condition_order : list of str
        Conditions in the right order

    Returns
    -------
    coil_cond_lst : list of strings containing the right conditions
    """
    drop_idx = []
    max_time_dif = 90000
    max_mep_dif = 7

    tm_pos_arr, tms_time_arr, tms_cur_arr, tms_idx_invalid = get_tms_elements(xml_paths)

    # get coil mean value
    tm_pos_arr = tm_pos_arr[2, :, :, :]

    # list for condition results
    conditions = []

    # index of instrument marker
    cond_idx = 0

    # iterate over all trigger marker
    for tm_index in range((tm_pos_arr.shape[0]) - 1):
        conditions.append(condition_order[cond_idx])
        if float(tms_cur_arr[tm_index]) == 0.:
            drop_idx.append(tm_index)
        tm_matrix_post = tm_pos_arr[tm_index + 1, :, :]
        tm_matrix = tm_pos_arr[tm_index, :, :]

        same_tm = arrays_similar(tm_matrix, tm_matrix_post)
        time_dif = tms_time_arr[tm_index + 1] - tms_time_arr[tm_index] > max_time_dif
        amp_dif = np.abs(float(tms_cur_arr[tm_index + 1]) - float(tms_cur_arr[tm_index])) > max_mep_dif
        if not same_tm and time_dif and amp_dif:
            arrays_similar(tm_matrix, tm_matrix_post)
            cond_idx += 1
            if cond_idx == len(condition_order):
                raise ValueError("Too many coil conditions found!")

    # assign last element to very last element
    conditions.append(conditions[-1])
    if cond_idx != len(condition_order) - 1:
        raise ValueError("Did not find all coil positions!")

    return conditions, drop_idx


def arrays_similar(tm_matrix, tm_matrix_post,  # , tm_mean_last,
                   pos_rtol=0, pos_atol=3.6, ang_rtol=.1, ang_atol=.1):
    """
    Compares angles and position for similarity.

    Splitting the comparison into angles and position is angebracht, as the absolute tolereance (atol) should be
    different for angles (degree) and position (millimeter) comparisons.

    Parameters:
    -----------
    tm_matrix: array-like, shape = (4,4)
        TMS Navigator triggermarker or instrument marker array

    tm_matrix_post: array-like, shape = (4,4)
        TMS Navigator triggermarker or instrument marker array

    tm_mean_last: array-like, shape = (4,4), optional
        Mean TMS Navigator triggermarker or instrument marker array for n zaps

    """
    last_pos = True
    last_ang = True
    # position
    pos = np.allclose(tm_matrix[0:3, 3], tm_matrix_post[0:3, 3], rtol=pos_rtol, atol=pos_atol)

    # angles
    ang = np.allclose(tm_matrix[0:3, 0:2], tm_matrix_post[0:3, 0:2], rtol=ang_rtol, atol=ang_atol)

    # if tm_mean_last is not None:
    #     last_pos = np.allclose(tm_matrix[0:3, 3], tm_mean_last[0:3, 3], rtol=pos_rtol, atol=pos_atol)
    #     last_ang = np.allclose(tm_matrix[0:3, 0:2], tm_mean_last[0:3, 0:2], rtol=ang_rtol, atol=ang_atol)

    next_same = pos and ang
    last_same = last_pos and last_ang
    return next_same


def match_tm_to_im(tm_array, im_array, tms_time_arr, tms_cur_arr):
    """
    Match triggermarker with instrumentmarker and get index of best fitting instrumentmarker.


    Parameters
    ----------
    tm_array : ndarray of float [Nx4x4]
        Mean-value of Nx4x4 coil matrices
    im_array : ndarray of float [Mx4x4]
        Instrument-marker-matrices

    Returns
    -------
    im_index_lst : list of int
        Indices of best fitting instrumentmarkers
    """
    max_time_dif = (tms_time_arr[1] - tms_time_arr[0]) * 3
    max_mep_dif = 9

    im_index_lst = []
    drop_idx = []

    for tm_index in range(tm_array.shape[0]):
        # after first zap, check time diff
        if tm_index > 0:
            if tms_time_arr[tm_index] - tms_time_arr[tm_index - 1] < max_time_dif:
                im_index_lst.append(im_index_lst[-1])
                continue

        allclose_index_lst = []
        diff_small = []

        atol_ori = 0.4
        atol_pos = 3
        repeat = False

        tm = tm_array[tm_index, :, :]

        # proc_diffs = np.argmin[procrustes(tm, im_array[i])[2] for i in range(im_array.shape[0])]

        # diffs = np.abs(tm - im_array)
        # diffs[0:3, 0:3] /= np.max(diffs[:, 0:3, 0:4], axis=0)
        # best_fit = np.argmin(np.array([np.sum(diffs[i]) for i in range(len(diffs))]))
        # small_diff_ori = int(np.argmin(np.array([np.sum(diffs[i][0:3, 0:3]) for i in range(len(diffs))])))
        # small_diff_pos = int(np.argmin(np.array([np.sum(diffs[i][0:3, 3]) for i in range(len(diffs))])))
        # a = rot_to_quat(tm[:3,:3])[1:]
        # b = [quaternion_diff(a, rot_to_quat(im_array[i, :3, :3])[1:]) for i in range(im_array.shape[0])]

        while not allclose_index_lst:
            if repeat:
                print(('Warning: Trigger marker #{:0>4}: No matching instrument marker within '
                       'atol_ori={} and atol_pos={}! Increasing tolerances by 0.1 and 0.5.'
                       .format(tm_index, atol_ori, atol_pos)))

                atol_ori = atol_ori + 0.1
                atol_pos = atol_pos + 0.5

            for im_index in range(im_array.shape[0]):

                # # check if arrays are close
                # if np.allclose(tm_array[tm_index, :, :], im_array[im_index, :, :], rtol=rtol):
                #     allclose_index_lst.append(im_index)

                # check if arrays are close
                diff = np.abs(tm_array[tm_index, :, :] - im_array[im_index, :, :])

                if (diff[0:3, 0:3] < atol_ori).all() and (diff[0:3, 3] < atol_pos).all():
                    diff_small.append(diff)
                    allclose_index_lst.append(im_index)

            if not allclose_index_lst:
                allclose_index_lst.append(-1)
                # repeat = True

        # if multiple arrays are close, choose value, with smallest difference
        if len(allclose_index_lst) > 1:
            smallest_value_index = int(np.argmin(np.array([np.sum(diff_small[i]) for i in range(len(diff_small))])))
            small_diff_ori = int(np.argmin(np.array([np.sum(diff_small[i][0:3, 0:3]) for i in range(len(diff_small))])))
            small_diff_pos = int(np.argmin(np.array([np.sum(diff_small[i][0:3, 3]) for i in range(len(diff_small))])))
            if not small_diff_ori == small_diff_pos:
                #     print allclose_index_lst
                print("Warning: Triggermarker #{:0>4}: " \
                      "Cannot decide for instrument marker , dropping this one. ".format(tm_index))
            #     drop_idx.append(tm_index)
            # im_index_lst.append(im_index_lst[-1])
            # else:
            # assert best_fit == allclose_index_lst[smallest_value_index]
            im_index_lst.append(allclose_index_lst[smallest_value_index])

        else:
            # assert best_fit == allclose_index_lst[0]
            im_index_lst.append(allclose_index_lst[0])

            # # if multile arrays are close, choose value, where the difference to the instrument marker has the smallest
            # # frobenius norm
            # if len(allclose_index_lst) > 1:
            #     smallest_value_index = int(np.argmin(np.linalg.norm(im_array[allclose_index_lst, :, :] -
            #                                                         tm_array[tm_index, :, :], axis=(1, 2))))
            #     im_index_lst.append(allclose_index_lst[smallest_value_index])
            # else:
            #     im_index_lst.append(allclose_index_lst[0])

    return im_index_lst, drop_idx


def get_instrument_marker(im_path):
    return get_marker(im_path, 'InstrumentMarker')[:2]


def get_marker(im_path, markertype):
    """
    Read instrument-marker and conditions from Neuronavigator .xml-file.

    Parameters
    ----------
    im_path : str or list of str
        Path to instrument-marker-file

    Returns
    -------
    im_array : np.array of float [Mx4x4]
        Instrument-marker-matrices
    im_cond_lst : list of str
        Labels of the instrument-marker-conditions
    im_marker_times : list of float
        Onset times
    """
    if isinstance(im_path, str):
        return get_single_marker_file(im_path, markertype)

    elif isinstance(im_path, list):
        # if multiple triggermarker files are provided, pick a marker with data
        im_array, marker_descriptions, marker_times = [], [], []

        # get data from all files
        for im in im_path:
            im_array_t, marker_descriptions, marker_times = get_single_marker_file(im, markertype)
            im_array.append(im_array_t)

        # get indices for all files where markers are empty
        empty_arr = []
        for arr in im_array:  # arr = im_array[0]
            empty_arr.append(markers_are_empty(arr))
        # assert np.all(np.sum(np.array(empty_arr).astype(int), axis=0) == 1)

        # go through the zaps and pick a marker with data.
        idx = []
        tmp = 0
        for i in range(len(im_array[0])):
            for j in range(len(im_array)):
                if not empty_arr[j][i]:
                    tmp = j
                # if all are empty, just use the last value (is empty anyway)
            idx.append(tmp)  # append

        # build marker idx based on idx
        final_arr = []
        for it, i in enumerate(idx):
            final_arr.append(im_array[i][it])
        return np.array(final_arr), marker_descriptions, marker_times
    else:
        raise NotImplementedError(f"type {type(im_path)} not implemented.")


def markers_are_empty(arr):
    return [marker_is_empty(arr[i, :, :]) for i in range(arr.shape[0])]


def marker_is_empty(arr):
    return np.all(arr == np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]))


def get_single_marker_file(im_path, markertype):
    """
        Read instrument-marker and conditions from Neuronavigator .xml-file.

        Parameters
        ----------
        im_path : str or list of str
            Path to instrument-marker-file

        Returns
        -------
        im_array : np.array of float [Mx4x4]
            Instrument-marker-matrices
        im_cond_lst : list of str
            Labels of the instrument-marker-conditions
        """
    im_array = np.empty([0, 4, 4])
    marker_descriptions = []
    marker_times = []
    # parse XML document
    im_tree = xmlt.parse(im_path)
    im_root = im_tree.getroot()
    # iterate over all 'InstrumentMarker' tags
    for marker_i in im_root.iter(markertype):
        marker_array = np.empty([1, 4, 4])
        # get tag were the matrix is
        if markertype == 'InstrumentMarker':
            marker_object = marker_i.find('Marker')
            marker_descriptions.append(marker_object.get('description'))
            matrix4d = marker_object.find('Matrix4D')

        elif markertype == 'TriggerMarker':
            matrix4d = marker_i.find('Matrix4D')
            marker_descriptions.append(marker_i.get('description'))
            marker_times.append(marker_i.get('recordingTime'))
        # get values
        for im_index1 in range(4):
            for im_index2 in range(4):
                marker_array[0, im_index1, im_index2] = (float(matrix4d.get('data' + str(im_index1) + str(im_index2))))
        im_array = np.append(im_array, marker_array, axis=0)

    return im_array, marker_descriptions, marker_times


def read_triggermarker_localite(fn_xml):
    """
    Read instrument-marker and conditions from neuronavigator .xml-file.

    Parameters
    ----------
    fn_xml : str
        Path to TriggerMarker.xml file
        (e.g. Subject_0/Sessions/Session_YYYYMMDDHHMMSS/TMSTrigger/TriggerMarkers_Coil1_YYYYMMDDHHMMSS.xml)

    Returns
    -------
    list of
        m_nnav : ndarray of float [4x4xN]
            Neuronavigation coordinates of N stimuli (4x4 matrices)
        didt : ndarray of float [N]
            Rate of change of coil current in (A/us)
        stim_int : ndarray of float [N]
            Stimulator intensity in (%)
        descr : list of str [N]
            Labels of the instrument-marker-conditions
        rec_time : list of str [N]
            Recording time in ms
    """

    m_nnav = np.empty([4, 4, 0])
    descr = []
    stim_int = []
    didt = []
    rec_time = []

    # parse XML document
    im_tree = xmlt.parse(fn_xml)
    im_root = im_tree.getroot()

    # iterate over all 'InstrumentMarker' tags
    for im_iter in im_root.iter('TriggerMarker'):
        m = np.empty([4, 4, 1])

        # read description
        descr.append(im_iter.get('description'))
        rec_time.append(im_iter.get('recordingTime'))

        # read di/dt and stimulator intensity
        im_rv = im_iter.find('ResponseValues').findall('Value')

        for _im_rv in im_rv:
            # di/dt
            if _im_rv.get('key') == "valueA":
                didt.append(float(_im_rv.get('response')))
            # stimulator intensity
            elif _im_rv.get('key') == "amplitudeA":
                stim_int.append(float(_im_rv.get('response')))

        # read matrix
        im_ma = im_iter.find('Matrix4D')

        for im_index1 in range(4):
            for im_index2 in range(4):
                m[im_index1, im_index2, 0] = (float(im_ma.get('data' + str(im_index1) + str(im_index2))))
        m_nnav = np.append(m_nnav, m, axis=2)

    didt = np.array(didt)
    stim_int = np.array(stim_int)

    return [m_nnav, didt, stim_int, descr, rec_time]



import numpy as np
import warnings


def ellipse_eccentricity( a , b ):
    """Calculates the eccentricity of an 2D ellipse with the
    semi axis a and b. An eccentricity of 0 corresponds to a sphere and an
    eccentricity of 1 means complete eccentric (line) with full restriction to
    the other axis

    Parameters
    ----------
    a : float
        First semi axis parameter
    b : float
        Second semi axis parameter

    Returns
    -------
    e : float
        Eccentricity (0...1)
    """
    return np.sqrt(1-(b/a)**2)


def rescale_lambda_centerized(D, s, tsc=False):
    """Rescales the eigenvalues of the matrix D
    according to their eccentricity. The scale factor is between 0...1 a
    scale factor of 0.5 would not alter the eigenvalues of the matrix D. A
    scale factor of 0 would unify all eigenvalues to one value such that it
    corresponds to a isotropic sphere. A scale factor of 1 alters the
    eigenvalues in such a way that the resulting ellipsoid is fully
    eccentric and anisotropic.

    Parameters
    ----------
    D : nparray of float [3 x 3]
        Diffusion tensor
    s : float
        Scale parameter [0 (iso) ... 0.5 (unaltered)... 1 (aniso)]
    tsc : boolean
        Tensor singularity correction

    Returns
    -------
    Ds : nparray of float [3 x 3]
        Scaled diffusion tensor
    """

    # print("Tensor scaling factor: " + str(s))
    # check if tensor matrix contains any element > 0
    if D.any():

        # reformat data if necessary
        reformat = 0

        if D.ndim == 1 and D.size == 9:
            D = np.reshape(D, (3, 3))
            reformat = 1

        # determine eigenvalues of original diffusion tensor matrix
        L, V = np.linalg.eig(D)
        l_idx = np.fliplr(np.argsort(L)[np.newaxis, :])[0]
        l_idx_reverse = np.argsort(l_idx)
        L = L[l_idx]
        V = V[:, l_idx]

        # calculate eccentricity e = [e12, e13, e23]
        e = ellipse_eccentricity(np.array([L[0], L[0], L[1]]), np.array([L[1], L[2], L[2]]))

        # calculate volume
        # Vol_old = 4.0/3 * np.pi * L[0]*L[1]*L[2]

        # scale eccentricity
        if s < 0.5:
            es = 2.0*e*s
        elif s > 0.5:
            es = 2.0*(1-e)*s+2*e-1
        elif s < 0:
            raise ValueError('ERROR: scale parameter s must not be smaller than 0!')
        elif s > 1:
            raise ValueError('ERROR: scale parameter s must not exceed 1!')
        elif s == 0.5:
            es = e
            warnings.warn('Scale parameter s = 0.5 ... diffusion tensor matrix is unaltered!')

        ls1 = 1
        ls2 = np.sqrt(1 - es[0]**2)
        ls3 = np.sqrt(1 - es[1]**2)

        Ls = np.array([ls1, ls2, ls3])

        # calculate scale factor for volume constraint
        scale = (L[0]*L[1]*L[2])/(Ls[0]*Ls[1]*Ls[2])

        # scale eigenvalues in order to fulfill volume constraint
        Ls = scale**(1.0/3)*Ls

        # compute scaled diffusion tensor matrix and reorder them
        Ds = np.dot(np.dot(V[:, l_idx_reverse], np.diag(Ls[l_idx_reverse])), V[:, l_idx_reverse].T)

        # reformat data back again to original format if input was like this
        if reformat:
            Ds = np.reshape(Ds, (9), order='F')

    elif tsc:
        Ds = np.array([1, 0, 0, 0, 1, 0, 0, 0, 1])

    else:
        Ds = D

    return Ds


def rescale_lambda_centerized_workhorse(D, s, tsc=False):
    """Rescales the eigenvalues of the matrix D
    according to their eccentricity. The scale factor is between 0...1 a
    scale factor of 0.5 would not alter the eigenvalues of the matrix D. A
    scale factor of 0 would unify all eigenvalues to one value such that it
    corresponds to a isotropic sphere. A scale factor of 1 alters the
    eigenvalues in such a way that the resulting ellipsoid is fully
    eccentric and anisotropic

    Parameters
    ----------
    D : ndarray of float [n x 9]
        Diffusion tensor
    s : float
        Scale parameter [0 (iso) ... 0.5 (unaltered)... 1 (aniso)]
    tsc : boolean
        Tensor singularity correction

    Returns
    -------
    Ds : list of nparray of float [3 x 3]
        Scaled diffusion tensor
    """

    Ds = np.empty(D.shape)

    for i in range(D.shape[0]):
        Ds[i, :] = rescale_lambda_centerized(D[i, :], s, tsc)

    return Ds



import re
import os
import sys
import h5py
import copy
import numpy as np
import multiprocessing


from pynibs import load_muaps
from multiprocessing import shared_memory
from .exp.Mep import sigmoid
from functools import partial
from pynibs import compute_chunks
from scipy.special import logsumexp


def DI_wave(t, intensity, t0=5, dt=1.4, width=0.25):
    """
    Determines cortical DI waves from TMS

    Parameters
    ----------
    t: ndarray of float [n_t]
        Time axis in ms
    intensity: float
        Stimulator intensity w.r.t resting motor threshold (typical range: [0 ... 2])
    t0: float
        offset time
    dt: float
        Spacing of waves in ms
    width: float
        Width of waves

    Returns
    -------
    y: ndarray of float [n_t]
        DI waves
    """

    waves = ["D", "I1", "I2", "I3", "I4"]

    x0 = dict()
    x0["D"] = 1.6952640144480995
    x0["I1"] = 1.314432218728424
    x0["I2"] = 1.4421623825084195
    x0["I3"] = 1.31643163560532
    x0["I4"] = 1.747079479469914

    amp = dict()
    amp["D"] = 12.83042571812661 / 35.46534715796085
    amp["I1"] = 35.46534715796085 / 35.46534715796085
    amp["I2"] = 26.15109003222628 / 35.46534715796085
    amp["I3"] = 15.491215097559184 / 35.46534715796085
    amp["I4"] = 10.461195366965226 / 35.46534715796085

    r = dict()
    r["D"] = 13.945868670402973
    r["I1"] = 8.707029476168504
    r["I2"] = 7.02266347578131
    r["I3"] = 16.74855628350182
    r["I4"] = 17.85806255278076

    y = np.zeros(len(t), dtype=np.float128)

    for i, w in enumerate(waves):
        y_ = np.exp(-(t-t0-i*dt)**2/(2*width**2))
        y_ = y_ / np.max(y_)
        y_ = y_ * sigmoid(intensity, amp=amp[w], r=r[w], x0=x0[w])
        y = y + y_

    return y


class Workhorse(object):

    def __init__(self, t, aMN, renshaw):
        self.t = t
        self.dt = t[1] - t[0]
        self.aMN = aMN
        self.renshaw = renshaw

    def run(self, input):
        """
        Workhorse to run alpha motor neurons in parallel

        Parameters
        ----------
        input: ndarray of float [n_t]
            Input signal to alpha motor neurons

        Returns
        -------
        alpha_motor_neurons: list of AlphaMotorNeuron instances
            Alpha motor neurons with calculated membrane potential (Vm) and spike times (spike_times)
        """

        # iterate over each time step
        feedback = np.zeros(len(self.t))
        delay = round(self.renshaw.delay / self.dt)

        for i, t_ in enumerate(self.t):
            feedback[i] = self.aMN.step(input, feedback, t_, i)

            if i + delay < len(self.t) - 1:
               feedback[i+1] = self.renshaw.step(feedback, t_, i)  # TODO consider delay and length of signal

        return self.aMN


class Model(object):
    """
    Spinal cord and Muscle model containing alpha motor neurons and Renshaw cells.
    """

    def __init__(self, T, dt, N_MU=100, fn_muaps="/data/pt_01756/studies/MEP_modeling/results/muscle_model"
                                                 "/Dist1_Monopolar_Abduct_NormalCV_New.hdf5", **kwargs):
        """
        Initialize spinal cord and Muscle model

        Parameters
        ----------
        T : Total Time of obesrvation
        dt : time step
        N_MU: Number of Motor units
        fn_muaps : Path to muap shapes
        **kwars: constants of neurons
        """

        # setup parameters and state variables
        self.N_MU = N_MU
        self.T = T
        self.dt = dt
        self.t = np.arange(0, T + dt, dt)

        # Neuron
        self.Vm_thr_array = kwargs['n_thr'] + kwargs['m_thr'] * np.arange(self.N_MU)  # thresholds of alpha MNs
        self.t_refrac = kwargs['t_refrac']
        self.tau_mem = kwargs['tau_mem']
        self.Vm_rest = kwargs['Vm_rest']
        self.Rm = kwargs['Rm']

        # Renshaw cell (inhibitory)
        self.slope_renshaw = kwargs.get('slope_renshaw')
        self.Vm_thr_renshaw = kwargs.get('Vm_thr_renshaw')
        self.Vm_rest_renshaw = kwargs.get('Vm_rest_renshaw')

        # Acetyl synapse (excitatory)
        self.tau_syn_decay_ac = kwargs.get('tau_syn_decay_ac')
        self.tau_syn_rise_ac = kwargs.get('tau_syn_rise_ac')
        self.tau_syn_decay_fast_ac = kwargs.get('tau_syn_decay_fast_ac')
        self.tau_syn_decay_slow_ac = kwargs.get('tau_syn_decay_slow_ac')
        self.a_syn_decay_fast_ac = kwargs.get('a_syn_decay_fast_ac')
        self.a_syn_decay_slow_ac = kwargs.get('a_syn_decay_slow_ac')

        # Glycin Synapse (inhibitory)
        self.tau_syn_decay_gl = kwargs.get('tau_syn_decay_gl')
        self.tau_syn_rise_gl = kwargs.get('tau_syn_rise_gl')
        self.tau_syn_decay_fast_gl = kwargs.get('tau_syn_decay_fast_gl')
        self.tau_syn_decay_slow_gl = kwargs.get('tau_syn_decay_slow_gl')
        self.a_syn_decay_fast_gl = kwargs.get('a_syn_decay_fast_gl')
        self.a_syn_decay_slow_gl = kwargs.get('a_syn_decay_slow_gl')

        # Connectivity
        self.connectivity_cortex_alpha  = kwargs.get('connectivity_cortex_alpha')
        self.connectivity_renshaw_alpha = kwargs.get('connectivity_renshaw_alpha')
        self.connectivity_alpha_renshaw = kwargs.get('connectivity_alpha_renshaw')

        # Load MUAPS
        self.fn_muaps = fn_muaps
        self.muaps = []
        self.t_muap = []
        self.muaps, t_muap = load_muaps(self.fn_muaps)

        # Workhorse Arrays
        self.alpha_motor_neurons = []
        self.renshaw_cells = []
        self.motor_units = []

        # Output Values
        self.mep = None
        self.mep_p2p = None

        # Synapse (Cortex -> Alpha Motor Neurons)
        acetyl_synapse_1 = Synapse(dt=self.dt, T=self.T,
                                   tau_syn_decay_fast=self.tau_syn_decay_fast_ac,
                                   tau_syn_decay_slow=self.tau_syn_decay_slow_ac,
                                   tau_syn_rise=self.tau_syn_rise_ac,
                                   a_syn_decay_slow=self.a_syn_decay_slow_ac,
                                   a_syn_decay_fast=self.a_syn_decay_fast_ac,
                                   connectivity=kwargs.get('connectivity_cortex_alpha'),
                                   N=self.N_MU)

        # Synapse (Alpha Motor Neurons -> Renshaw Cells)
        acetyl_synapse_2 = Synapse(dt=self.dt, T=self.T,
                                   tau_syn_decay_fast=self.tau_syn_decay_fast_ac,
                                   tau_syn_decay_slow=self.tau_syn_decay_slow_ac,
                                   tau_syn_rise=self.tau_syn_rise_ac,
                                   a_syn_decay_slow=self.a_syn_decay_slow_ac,
                                   a_syn_decay_fast=self.a_syn_decay_fast_ac,
                                   connectivity=self.connectivity_alpha_renshaw,
                                   N=self.N_MU)

        # Synapse (Renshaw Cells -> Alpha Motor Neurons)
        glycine_synapse = Synapse(dt=self.dt, T=self.T,
                                  tau_syn_decay_fast=self.tau_syn_decay_fast_gl,
                                  tau_syn_decay_slow=self.tau_syn_decay_slow_gl,
                                  tau_syn_rise=self.tau_syn_rise_gl,
                                  a_syn_decay_slow=self.a_syn_decay_slow_gl,
                                  a_syn_decay_fast=self.a_syn_decay_fast_gl,
                                  connectivity=self.connectivity_renshaw_alpha,
                                  N=self.N_MU)

        self.aMNs = AlphaMotorNeuron(tau_mem=self.tau_mem, Vm_thr=self.Vm_thr_array, Vm_rest=self.Vm_rest, Rm=self.Rm,
                                     t_refrac=self.t_refrac, T=self.T, dt=self.dt,
                                     synapses=[acetyl_synapse_1, glycine_synapse], N=self.N_MU)

        self.renshaw = RenshawCell(tau_mem=self.tau_mem, Vm_thr=self.Vm_thr_renshaw, Vm_rest=self.Vm_rest, Rm=self.Rm,
                                   T=self.T, dt=self.dt, delay=0,
                                   synapses=[acetyl_synapse_2],
                                   slope=self.slope_renshaw)  # TODO own Vm_thr for optimisation

        for i in range(self.N_MU):
            self.motor_units.append(MotorUnit(muap=self.muaps[:, i], t=self.t))

    def run(self, inp):
        """
        Starts the parallel processing.
        Creates Model can all workhorse process

        Parameters
        ----------
        inp: ndarray of float [n_t]
            Input signal to alpha motor neurons
        """

        workhorse = Workhorse(self.t, self.aMNs, self.renshaw)
        self.alpha_motor_neurons = workhorse.run(inp)

        muscle = Muscle(self.alpha_motor_neurons, self.motor_units)
        self.mep = muscle.get_mep()
        self.mep_p2p = np.max(self.mep) - np.min(self.mep)


class Synapse(object):
    """
    Synapse (conductance-based)
    """

    def __init__(self, T, dt, N, tau_syn_decay=None, tau_syn_rise=None,
                 tau_syn_decay_fast=None, tau_syn_decay_slow=None, a_syn_decay_fast=None, a_syn_decay_slow=None,
                 connectivity=1):
        """
        Initializes Synapse instance

        Parameters
        ----------
        T : float
            Total simulation time in ms
        dt : float
            Time-step to solve DEQ
        tau_syn_decay : float
            Decay time constant of synapse in ms (set this parameter, if only one decay time constant is present and set
            tau_syn_decay_fast=None, tau_syn_decay_slow=None)
        tau_syn_rise : float
            Rise time of synapse in ms
        tau_syn_decay_fast : float
            Decay time constant of synapse of fast component in ms
        tau_syn_decay_slow : float
            Decay time constant of synapse of slow component in ms
        a_syn_decay_fast : float
            Proportion of fast decay time constant
        a_syn_decay_slow : float
            Proportion of slow decay time constant
        connectivity : float
            Scaling factor for synaptic strength
        """

        self.tau_syn_decay = tau_syn_decay
        self.T = T
        self.dt = dt
        self.connectivity = connectivity
        self.tau_syn_rise = tau_syn_rise
        self.tau_syn_decay_fast = tau_syn_decay_fast
        self.tau_syn_decay_slow = tau_syn_decay_slow
        self.a_syn_decay_fast = a_syn_decay_fast
        self.a_syn_decay_slow = a_syn_decay_slow
        self.synapse_single_decay = False
        self.synapse_double_decay = False
        self.synapse_single_decay_rise = False
        self.synapse_double_decay_rise = False

        if tau_syn_decay is not None and (
                tau_syn_decay_slow is None and tau_syn_decay_fast is None and tau_syn_rise is None):
            self.synapse_single_decay = True

        if (tau_syn_decay_slow is not None and tau_syn_decay_fast is not None) and (
                tau_syn_rise is None and tau_syn_decay is None):
            self.synapse_double_decay = True

        if (tau_syn_decay_slow is not None and tau_syn_decay_fast is not None and tau_syn_rise is not None) and (
                tau_syn_decay is None):
            self.synapse_double_decay_rise = True

        if (tau_syn_rise is not None and tau_syn_decay is not None) and (
                tau_syn_decay_slow is None and tau_syn_decay_fast is None):
            self.synapse_single_decay_rise = True

        if self.synapse_single_decay + self.synapse_double_decay + self.synapse_single_decay_rise + self.synapse_double_decay_rise != 1:
            raise AssertionError("Please check assignment of time constants...")

        # initialize arrays
        self.t = np.linspace(0, self.T, int(self.T / self.dt + 1))
        self.g_f = np.zeros((N, len(self.t)))
        self.x_f = np.zeros((N, len(self.t)))
        self.g_s = np.zeros((N, len(self.t)))
        self.x_s = np.zeros((N, len(self.t)))
        self.x = np.zeros((N, len(self.t)))
        self.g_tot = np.zeros((N, len(self.t)))

    def step(self, inp, i):
        """
        Solves the DEQ and computes synapse conductivity

        Parameters
        ----------
        inp: ndarray of float [n_t]
            Input signal
        i: int
            Current index
        """

        # double exponential kernel with rise time and slow + fast decay times
        if self.synapse_double_decay_rise:

            self.g_f[:, i] = self.g_f[:, i - 1] + self.x_f[:, i - 1] * self.dt
            self.x_f[:, i] = self.x_f[:, i - 1] + (
                    (inp[i] - self.x_f[:, i - 1]) * (self.tau_syn_rise + self.tau_syn_decay_fast) -
                    self.g_f[:, i - 1]) / (self.tau_syn_rise * self.tau_syn_decay_fast) * self.dt
            self.g_s[:, i] = self.g_s[:, i - 1] + self.x_s[:, i - 1] * self.dt
            self.x_s[:, i] = self.x_s[:, i - 1] + (
                    (inp[i] - self.x_s[:, i - 1]) * (self.tau_syn_rise + self.tau_syn_decay_slow) -
                    self.g_s[:, i - 1]) / (self.tau_syn_rise * self.tau_syn_decay_slow) * self.dt
            self.g_tot[:, i] = self.a_syn_decay_fast * self.g_f[:, i] + self.a_syn_decay_slow * self.g_s[:, i]
            
        # single exponential kernel with slow + fast decay times
        if self.synapse_double_decay:
            self.g_s[:, i] = self.g_s[:, i - 1] + ((0 - self.g_s[:, i - 1]) / self.tau_syn_decay_slow) * self.dt + inp[
                i]
            self.g_f[:, i] = self.g_f[:, i - 1] + ((0 - self.g_f[:, i - 1]) / self.tau_syn_decay_fast) * self.dt + inp[
                i]
            self.g_tot[:, i] = self.a_syn_decay_fast * self.g_f[:, i] + self.a_syn_decay_slow * self.g_s[:, i]

        # single exponential kernel with one decay time
        if self.synapse_single_decay:
            self.g_tot[:, i] = self.g_tot[:, i - 1] + ((0 - self.g_tot[:, i - 1]) / self.tau_syn_decay) * self.dt + inp[
                i]

        # double exponential kernel with rise time and one decay time
        if self.synapse_single_decay_rise:
            self.g_tot[:, i] = self.g_tot[:, i - 1] + self.x[:, i - 1] * self.dt
            self.x[:, i] = self.x[:, i - 1] + ((inp[i] - self.x[:, i - 1]) * (self.tau_syn_rise + self.tau_syn_decay) -
                                               self.g_tot[:, i - 1]) / (
                                       self.tau_syn_rise * self.tau_syn_decay) * self.dt

        return self.g_tot[:, i] * self.connectivity  # Apply weight to output


class RenshawCell(object):
    """
    Renshaw Cell (Leaky integrate and fire)
    """

    def __init__(self, tau_mem, Vm_thr, Vm_rest, Rm, T, dt, delay, synapses, slope):
        """
        Initializes Renshaw Cell instance

        Parameters
        ----------
        tau_mem : float
            Time constant of cell membrane in ms
        Vm_thr : float
            Firing threshold in mV
        Vm_rest : float
            Resting potential
        Rm : float
            Cell membrane resistance
        T : float
            Total simulation time in ms
        dt : float
            Time-step to solve DEQ
        delay: float
            Delay of feedback in ms
        synapses: list of Synapses
            List of all synapses
        slope : float
            Slope of Threshold
        """
        self.tau_mem = tau_mem
        self.Vm_thr = Vm_thr  #TODO proper Threshold
        self.Vm_rest = Vm_rest
        self.Rm = Rm

        self.T = T
        self.dt = dt
        self.t_rest = 0

        self.max_firerate = 23.12  #TODO Proper value
        self.slope = slope
        self.delay = delay

        # Synapses
        self.acetyl_synapse = synapses[0]  # excitatory (Alpha Motor Neuron)

        # initialize arrays
        self.t = np.linspace(0, self.T, int(self.T / self.dt + 1))
        self.g_tot = np.zeros(len(self.t))
        self.Vm = Vm_rest * np.ones(len(self.t),dtype=np.float128)
        self.spike_times = np.zeros(len(self.t))
        self.output = np.zeros(len(self.t),dtype=np.float128)

    def step(self, inp, t_, i):
        """
        Solves the DEQ and computes membrane potential.
        Saves membrane potential in self.Vm and spike times in self.spike_times

        Parameters
        ----------
        inp: ndarray of float [n_t]
            Input signal
        t_ : floatf
            Current time step
        i : int
            Current index
        """

        self.g_tot[i] = np.sum(self.acetyl_synapse.step(inp, i))

        self.Vm[i] = self.Vm[i - 1] + (-(self.Vm[i - 1] - self.Vm_rest) - self.Vm[i - 1] * self.g_tot[i] * self.Rm)/self.tau_mem * self.dt

        #A. Spiegler potential to Rate
        #np.seterr('raise')
        self.output[i] = 2*self.max_firerate / (1+(logsumexp(self.slope*(self.Vm_thr-self.Vm[i]))))

        return self.output[i]

        # if(self.Vm[i] >= self.Vm_thr):
        #     self.spike_times[i] = t_
        #     self.Vm[i] = self.Vm_rest
        #     self.t_rest = t_ + self.t_refrac
        #     return 1 / self.dt #TODO in Feuerate umwandeln
        #
        # return 0  # return no spike


class AlphaMotorNeuron(object):
    """
    Alpha Motor Neuron (Leaky integrate and fire)
    """

    def __init__(self, tau_mem, Vm_thr, Vm_rest, Rm, t_refrac, T, dt, synapses, N):
        """
        Initializes AlphaMotorNeuron instance

        Parameters
        ----------
        tau_mem : float
            Time constant of cell membrane in ms
        Vm_thr : float
            Firing threshold in mV
        Vm_rest : float
            Resting potential
        Rm : float
            Cell membrane resistance
        t_refrac : float
            Refractory time
        T : float
            Total simulation time in ms
        dt : float
            Time-step to solve DEQ
        synapses: list of Synapses
            List of all synapses
        N: int
            Number of alpha motor neurons
        """
        self.tau_mem = tau_mem
        self.Vm_thr = Vm_thr
        self.Vm_rest = Vm_rest
        self.Rm = Rm
        self.t_refrac = t_refrac
        self.T = T
        self.N = N
        self.dt = dt
        self.t_rest = np.zeros(N)
        self.is_in_refrac = np.ones(N)

        # Synapses
        self.acetyl_synapse = synapses[0]  # excitatory (from motor cortex)
        self.glycin_synapse = synapses[1]  # inhibitory (from Renshaw cell)

        # initialize arrays
        self.t = np.linspace(0, self.T, int(self.T / self.dt + 1))
        self.g_tot = np.zeros((N, len(self.t)))
        self.Vm = Vm_rest * np.ones((N, len(self.t)), dtype=np.float128)
        self.spike_times = np.zeros((N, len(self.t)))

        self.anySpike = 0

    def step(self, inp, renshaw_feedback, t_, i):
        """
        Solves the DEQ and computes membrane potential.
        Saves membrane potential in self.Vm and spike times in self.spike_times

        Parameters
        ----------
        inp: ndarray of float [n_t]
            Input signal
        renshaw_feedback: ndarray of float
            Feedback from Renshaw Cell
        t_ : float
            Current time
        i : int
            Current index
        """
        x = 0  # self.glycin_synapse.step(renshaw_feedback, i)
        y = self.acetyl_synapse.step(inp, i)

        self.g_tot[:, i] = np.where(x+y < 0, 0, x+y)

        # if t_ > self.t_rest:
        self.Vm[:, i] = self.Vm[:, i - 1] + self.is_in_refrac * \
                        ((-(self.Vm[:, i - 1] - self.Vm_rest) - self.Vm[:, i - 1] * self.g_tot[:, i] * self.Rm) /
                         self.tau_mem * self.dt)
        firing_neurons = zip(*np.where(self.Vm[:, i] >= self.Vm_thr))

        for f in firing_neurons:
            if self.is_in_refrac[f] == 1:
                self.spike_times[f, i] = t_
                self.Vm[f, i] = self.Vm_rest
                self.t_rest[f] = t_ + self.t_refrac
                self.is_in_refrac[f] = 0
                self.anySpike = 1

        for k, t in enumerate(self.t_rest):
            if t_ > self.t_rest[k]:
                self.is_in_refrac[k] = 1

        # spike event (reset to Vm_rest and pause integrating during refractory period)
        if self.anySpike == 1:
            self.anySpike = 0
            return 1 / self.dt

        # if np.any(self.Vm[:, i] >= self.Vm_thr):
        #     # check if still in refract
        #     np.where(t_ > self.t_rest and self.Vm[:, i] >= self.Vm_thr, self.is_in_refrac, self.is_in_refrac + 1)
        #     print((self.is_in_refrac))
        #     self.spike_times.append(t_ * self.is_in_refrac)  # if in refrac append 0
        #     self.Vm[:, i] = self.Vm_rest * (1 - self.is_in_refrac) + self.Vm[:, i]  # TODO
        #     self.t_rest = (t_ + self.t_refrac) * (1 - self.is_in_refrac) + self.t_rest * self.is_in_refrac
        #
        #     np.where(self.Vm[:, i] >= self.Vm_thr, self.is_in_refrac,
        #              self.is_in_refrac * 0)  # Sets all Amn 0 when in refrac
        #
        #     return 1 / self.dt  # return spike

        return 0  # return no spike


class MotorUnit(object):
    """
    Motor unit
    """

    def __init__(self, muap, t):
        """
        Initializes motor unit instance

        Parameters
        ----------
        muap: ndarray of float [n_t]
            Motor unit action potential
        t: ndarray of float [n_t]
            Time axis corresponding to motor unit action potential
        t_total: ndarray of float [n_t]
            Total time axis of observation (has to fit to t)
        """

        self.muap = muap
        self.t = t

    def get_muap(self, spike_times, k):
        """
        Adds MUAPs of the motorunit at given spiketimes

        Parameters
        ----------
        spike_times: ndarray of float [n_spikes]
            Spike times

        Returns
        -------
        muap_total: ndarray of float [n_t]
            Total MUAP signal (MUAPs are summed at spiketimes)
        """
        muap_total = np.zeros(len(self.t))

        for tau in spike_times[k, :]:
            if not tau == 0:
                try:
                    idx_tau = np.argmin(np.abs(self.t - tau))
                    if len(muap_total) - idx_tau < len(self.muap):
                        muap_total[idx_tau:idx_tau + len(self.muap)] += self.muap[:len(muap_total) - idx_tau]
                    else:
                        muap_total[idx_tau:idx_tau + len(self.muap)] += self.muap

                except Exception as error:
                    print(error)
        return muap_total


class Muscle(object):
    """
    Muscle containing multiple motor unit instances
    """

    def __init__(self, alpha_motor_neurons, motor_units):
        """
        Initializes muscle instance

        Parameters
        ----------
        alpha_motor_neurons: list of AlphaMotorNeuron instances
            AlphaMotorNeuron instances of the muscle
        motor_units: list of MotorUnit instances
            MotorUnit instances corresponding to the AlphaMotorNeuron instances
        """

        self.alpha_motor_neurons = alpha_motor_neurons
        self.motor_units = motor_units
        self.t = self.alpha_motor_neurons.t
        self.N = self.alpha_motor_neurons.N
        self.mep = np.zeros(len(self.t))

        # import matplotlib.pyplot as plt
        # plt.plot(self.t, alpha_motor_neurons.Vm[80, :])
        # plt.title("80-Alpha-Motorneuron")
        # plt.xlabel("t[ms]")
        # plt.ylabel("PSP [mV]")
        # plt.show()

    def get_mep(self):
        for k in range(self.N):
           self.mep += self.motor_units[k].get_muap(self.alpha_motor_neurons.spike_times, k)

        return self.mep

# def spindle(self):
#
#     import matplotlib.pyplot as plt
#     fig,ax = plt.subplots()
# ax.plot(self.t,self.inp)
# hz_500_signal = np.arange(30,self.T,2) #Delay auf 20-50ms
# print(hz_500_signal)
# print(self.t)
# self.inp[np.nonzero(hz_500_signal[:, None] == self.t)[1]] = 1
# ax.plot(self.t,self.inp)
# plt.show()

# Paper Values Lin 2002

# out = np.zeros(len(self.t)+1)
# self.g_g = 40 # impulse/s
# self.g_f = 4 # N
# self.f_me = np.ones(len(self.t)+1)*0.00001 # N (random choosen to test)
# for i in range(len(self.t)):
#     inp = self.g_g + np.log(self.f_me / self.g_f + 1)
#
#     if i == 0:
#         A = 40 * ((1.7 / self.dt ** 2) * (inp[i + 1] - 2 * inp[i] +0) + (
#                     2.58 / (2 * self.dt) * (inp[i + 1] - 0)) + 0.4 * inp[i])
#         B = (-2.2 / (2 * self.dt) * (out[i + 1] - 0)) - (1 / self.dt ** 2) * out[i + 1] + (
#                     -1 / self.dt ** 2) * 0
#     else:
#         A = 40 * ((1.7 / self.dt ** 2) * (inp[i + 1] - 2 * inp[i] + inp[i - 1]) + (
#                     2.58 / (2 * self.dt) * (inp[i + 1] - inp[i - 1])) + 0.4 * inp[i])
#         B = (-2.2 / (2 * self.dt) * (out[i + 1] - out[i - 1])) - (1 / self.dt ** 2) * out[i + 1] + (
#                     -1 / self.dt ** 2) * out[i - 1]
#
#
#     out[i] = ((-2/self.dt)+0.4)*(A+B)

# ax.plot(out)
# plt.show()


# N=0, c1=48144, c2=45845, alpha1=0.0326, alpha2=0.034,
# a=220, b_h=10, K_sh=35, K_ph=5

# self.t = t
# # Values from Patet2015 Paper
# self.c1 = c1
# self.c2 = c2
# self.alpha1 = alpha1
# self.alpha2 = alpha2
# self.T0 = (t[1]-t[0])#/1000 # Has to be converted to seconds
#
# self.a = a  # Hz/cm
# self.b_h = b_h  # g*s/cm
# self.K_sh = K_sh  # g/cm spindle spring constants
# self.K_ph = K_ph  # g/cm spindle spring constants TODO find suitable values
# self.beta = (self.K_ph + self.K_sh) / self.b_h
#
# # Values from Patet2015 Paper
#
# self.w = np.zeros(len(t))
# self.active_f = np.zeros(len(t))
# self.spindle_stretch = np.zeros(len(t))
# #self.spindle_stretch[np.flatnonzero(t == 50)[0]: np.flatnonzero(t == 150)[0]] = 0.1  # TODO Test value proberly implement it (simulation ?)
# self.spindle_stretch[50:200] = 0.1
# self.S_Ia = np.zeros(len(t))
# self.S_II = np.zeros(len(t))
#
# self.number_spikes = 0
#
# #Only used to calculated the MEP
# self.muap = muap
# self.spike_times = spike_times
# self.N = N

# def spindle(self,i,spikes_t):
#
#     if not i == len(self.t) - 1:
#         # TODO Move to Muscle
#
#         #for _, sp in enumerate(spikes_t):
#            # print(self.t[i])
#         sp = spikes_t[-1]
#
#         self.active_f[i] += self.c1 * np.exp(-(1 / self.alpha1) * (self.t[i] - sp)) - self.c2 * np.exp(
#             -(1 / self.alpha2) * (self.t[i] - sp)) # Rember to add *1/1000 to the exponent
#
#
#         self.Bh1 = -((self.K_sh ** 2 * (np.exp(self.beta * self.T0) - 1)) / (
#                     (self.K_ph + self.K_sh) * np.exp(self.beta * self.T0)))
#         self.Bh2 = ((self.K_sh * (np.exp(self.beta * self.T0) - 1)) / (
#                     (self.K_ph + self.K_sh) * np.exp(self.beta * self.T0)))
#         test = np.array([self.Bh1, self.Bh2], dtype=np.float64)
#         test2 = np.array([self.spindle_stretch[i], self.active_f[i]], dtype=np.float64)
#         self.w[i+1] = (self.w[i] + self.Bh1*self.spindle_stretch[i]+self.Bh2 * self.active_f[i])*(np.exp(-self.beta*self.T0))
#
#         self.S_Ia[i] = self.a / self.K_sh * self.w[i] + self.a * self.spindle_stretch[i]
#         self.S_II[i] = -self.a / self.K_sh * self.w[i]
#
#     return self.w, self.active_f,self.spindle_stretch,self.S_Ia,self.S_II



import os
import h5py
from .regression import *
import scipy
import pynibs
import numpy as np
import pandas as pd
from .exp.Mep import linear
from .exp.Mep import sigmoid
from .hdf5_io import write_geo_hdf5
from .hdf5_io import write_temporal_xdmf
from .hdf5_io import load_mesh_hdf5
import multiprocessing.pool
import matplotlib.pyplot as plt
from _functools import partial
from sklearn.neighbors import KernelDensity


# Workhorses for multiprocessing
###################################################################
def workhorse_mc(idx_list, array, ele_idx_1, mode="cols"):
    """
    Determines mutual coherence for given zap indices in idx_list.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    ele_idx_1 : ndarray of float [n_ele]
        Element indices for which the optimization is performed
    mode : str, optional, default: "cols"
        Set if the mutual coherence is calculated w.r.t. columns or rows ("cols", "rows")

    Returns
    -------
    res : ndarray of float [n_combs]
        Mutual coherence. Lower values indicate more orthogonal e-field combinations (better)
    """

    res = []
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    for ind in idx_list:
        if mode == "cols":
            res.append(pynibs.mutual_coherence(array[ind, :]))
        elif mode == "rows":
            res.append(pynibs.mutual_coherence(array[ind, :].transpose()))
        else:
            raise NotImplementedError("Specified mode not implemented. Choose 'rows' or 'cols'.")

    return res


# TODO Benjamin
def workhorse_corr(idx_list, array, ele_idx_1):
    """

    :param idx_list:
    :param array:
    :param ele_idx_1:
    :return:
    """
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        r = np.corrcoef(array[ind, :])
        res.append(np.mean(r[np.triu_indices(r.shape[0], k=1)]))

    return res


def workhorse_var(idx_list, array, ele_idx_1):
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        res.append(np.mean(np.var(array[ind, :], axis=0)))

    return res


def workhorse_smooth(idx_list, array, ele_idx_1):
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        res.append(np.var(np.mean(array[ind, :], axis=0)))

    return res


def workhorse_svd(idx_list, array, ele_idx_1):
    """
    Determines condition number for given zap indices in idx_list.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    ele_idx_1 : ndarray of float [n_ele]
        Element indices for which the optimization is performed

    Returns
    -------
    res : ndarray of float [n_combs]
        Condition number. Lower values indicate more orthogonal e-field combinations (better)
    """
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        s = scipy.linalg.svd(array[ind, :], compute_uv=False)
        res.append(np.max(s)/np.min(s))

    return res


def workhorse_variability(idx_list, array, ele_idx_1):
    """
    Determines variability score for given zap indices in idx_list.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    ele_idx_1 : ndarray of float [n_ele]
        Element indices for which the optimization is performed

    Returns
    -------
    res : ndarray of float [n_combs]
        Condition number. Lower values indicate more orthogonal e-field combinations (better)
    """
    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    res = []

    for ind in idx_list:
        D = np.zeros((array.shape[1], array.shape[1]))
        d = 0

        for col_idx in range(0, array.shape[1]-1):
            for row_idx in range(col_idx+1, array.shape[1]):
                D[row_idx, col_idx] = np.linalg.norm(array[ind, row_idx] - array[ind, col_idx])
                d += D[row_idx, col_idx]

        res.append(1/d)

    return res


def workhorse_dist(idx_list, array, ele_idx_1):
    """
    Determines distance score for given zap indices in idx_list.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    ele_idx_1 : ndarray of float [n_ele]
        Element indices for which the optimization is performed

    Returns
    -------
    res : ndarray of float [n_combs]
        Distance based score. Lower values indicate more equidistant sampling (better)
    """
    res = np.zeros(len(idx_list))

    array_dist = array[:, ele_idx_1]

    if array_dist.ndim == 1:
        array_dist = array_dist[:, np.newaxis]

    e_max = np.max(array_dist, axis=0)
    e_min = np.min(array_dist, axis=0)

    for j, ind in enumerate(idx_list):
        p = np.vstack((e_min, array_dist[ind, :], e_max))
        d_var = np.zeros(array_dist.shape[1])

        for i in range(array_dist.shape[1]):
            p_sort = np.sort(p[:, i])

            if p_sort[0] == p_sort[1]:
                p_sort = p_sort[1:]

            if p_sort[-2] == p_sort[-1]:
                p_sort = p_sort[:-1]

            d_var[i] = np.var(np.diff(p_sort))

        res[j] = np.mean(d_var)

    return res


def workhorse_dist_svd(idx_list, array, ele_idx_1, ele_idx_2):
    """
    Determines distance score and condition number for given zap indices in idx_list. If c_max_idx is given,
    the distance based score is calculated only for this element.
    The condition number however is optimized for all elements in array

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    ele_idx_1 : ndarray of float [n_ele]
        Element indices for which the dist optimization is performed for
    ele_idx_2 : ndarray of float [n_ele]
        Element indices for which the svd optimization is performed for

    Returns
    -------
    res_dist : ndarray of float [n_combs]
        Distance based score. Lower values indicate more equidistant sampling (better)
    res_svd : ndarray of float [n_combs]
        Condition number. Lower values indicate more orthogonal e-field combinations (better)
    """
    res_dist = np.zeros(len(idx_list))
    res_svd = np.zeros(len(idx_list))

    array_dist = array[:, ele_idx_1]

    if array_dist.ndim == 1:
        array_dist = array_dist[:, np.newaxis]

    array_svd = array[:, ele_idx_2]

    if array_svd.ndim == 1:
        array_svd = array_svd[:, np.newaxis]

    e_max = np.max(array_dist, axis=0)
    e_min = np.min(array_dist, axis=0)

    for j, ind in enumerate(idx_list):

        # svd
        u, s, vh = scipy.linalg.svd(array_svd[ind, :])
        res_svd[j] = np.max(s)/np.min(s)

        # dist
        p = np.vstack((e_min, array_dist[ind, :], e_max))
        d_var = np.zeros(array_dist.shape[1])

        for i in range(array_dist.shape[1]):
            p_sort = np.sort(p[:, i])

            if p_sort[0] == p_sort[1]:
                p_sort = p_sort[1:]

            if p_sort[-2] == p_sort[-1]:
                p_sort = p_sort[:-1]

            d_var[i] = np.var(np.diff(p_sort))

        res_dist[j] = np.mean(d_var)

    return res_dist, res_svd


def workhorse_dist_mc(idx_list, array, ele_idx_1, ele_idx_2, mode="cols"):
    """
    Determines distance score and mutual coherence for given zap indices in idx_list. If c_max_idx is given,
    the distance based score is calculated only for this element.
    The condition number however is optimized for all elements in array

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    mode : str, optional, default: "cols"
        Set if the mutual coherence is calculated w.r.t. columns or rows ("cols", "rows")
    ele_idx_1 : ndarray of float [n_ele]
        Element indices for which the dist optimization is performed for
    ele_idx_2 : ndarray of float [n_ele]
        Element indices for which the mc optimization is performed for

    Returns
    -------
    res_dist : ndarray of float [n_combs]
        Distance based score. Lower values indicate more equidistant sampling (better)
    res_mc : ndarray of float [n_combs]
        Mutual coherence. Lower values indicate more orthogonal e-field combinations (better)
    """
    res_dist = np.zeros(len(idx_list))
    res_mc = np.zeros(len(idx_list))

    array_dist = array[:, ele_idx_1]

    if array_dist.ndim == 1:
        array_dist = array_dist[:, np.newaxis]

    array_mc = array[:, ele_idx_2]

    if array_mc.ndim == 1:
        array_mc = array_mc[:, np.newaxis]

    e_max = np.max(array_dist, axis=0)
    e_min = np.min(array_dist, axis=0)

    for j, ind in enumerate(idx_list):

        # mc
        if mode == "cols":
            res_mc[j] = pynibs.mutual_coherence(array_mc[ind, :])
        elif mode == "rows":
            res_mc[j] = pynibs.mutual_coherence(array_mc[ind, :].transpose())
        else:
            raise NotImplementedError("Specified mode not implemented. Choose 'rows' or 'cols'.")

        # dist
        p = np.vstack((e_min, array_dist[ind, :], e_max))
        d_var = np.zeros(array_dist.shape[1])

        for i in range(array_dist.shape[1]):
            p_sort = np.sort(p[:, i])

            if p_sort[0] == p_sort[1]:
                p_sort = p_sort[1:]

            if p_sort[-2] == p_sort[-1]:
                p_sort = p_sort[:-1]

            d_var[i] = np.var(np.diff(p_sort))

        res_dist[j] = np.mean(d_var)

    return res_dist, res_mc


def workhorse_coverage_prepare(idx_list, array, zap_idx):
    """Prepares coverage calculation.
    Determines coverage distributions for elements in idx_list given the zaps in zap_idx

    Parameters
    ----------
    idx_list : list [n_ele]
        Index lists of elements.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    zap_idx : ndarray of int
        Included zaps in coverage distribution.

    Returns
    -------
    x : ndarray of float [200 x n_ele]
        x-values of coverage distributions, defined in interval [0, 1] (element wise normalized electric field)
    y : ndarray of float [200 x n_ele]
        y-values of coverage distributions (element wise probability of already included e-fields)
    """

    n_x = 200

    x = np.zeros((n_x, len(idx_list)))
    y = np.zeros((n_x, len(idx_list)))

    kde = KernelDensity(bandwidth=0.03, kernel='gaussian')

    for j, ind in enumerate(idx_list):
        e_min = np.min(array[:, ind])
        e_max = np.max(array[:, ind])
        e_samples = (array[zap_idx, ind] - e_min) / (e_max - e_min)

        if not isinstance(e_samples, np.ndarray):
            e_samples = np.array([e_samples])

        kde_ele = kde.fit(e_samples[:, np.newaxis])
        x[:, j] = np.linspace(0, 1, n_x)
        y[:, j] = np.exp(kde_ele.score_samples(x[:, j][:, np.newaxis]))

    return x, y


def workhorse_coverage(idx_list, array, x, y, ele_idx_1):
    """
    Determine coverage score (likelihood) for given zap indices in idx_list

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    x : ndarray of float [200 x n_ele]
        x-values of coverage distributions, defined in interval [0, 1] (element wise normalized electric field)
    y : ndarray of float [200 x n_ele]
        y-values of coverage distributions (element wise probability of already included e-fields)
    ele_idx_1 : ndarray of float [n_roi]
        Element indices for which the coverage optimization is performed for

    Returns
    -------
    res : ndarray of float [n_combs]
        Coverage score (likelihood) for given electric field combinations. Lower values indicate that the
        new zap fills a gap which was not covered before.
    """

    p_e = np.zeros(len(idx_list))

    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    for j, ind in enumerate(idx_list):
        p_e_zap = np.zeros(array.shape[1])

        # normalized e-fields of this zap in all elements
        e_zap = (array[ind[-1], :] - np.min(array, axis=0)) / (np.max(array, axis=0) - np.min(array, axis=0))

        # determine e-field coverage probability for every element
        for i_ele in range(array.shape[1]):
            p_e_zap[i_ele] = np.interp(x=e_zap[i_ele], xp=x[:, i_ele], fp=y[:, i_ele])

        # accumulate e-field coverage over ever elements using log-likelihood
        p_e_zap[p_e_zap <= 0] = 1e-100
        p_e[j] = np.sum(p_e_zap)

    return p_e


def workhorse_fim(idx_list, array, ele_idx_1, e_opt, c=None):
    """
    Determine difference between e-fields and optimal e-field determined using the Fisher Information Matrix.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    ele_idx_1 : ndarray of float [n_roi]
        Element indices for which the fim optimization is performed for
    e_opt : ndarray of float [n_roi]
        Optimal electric field value(s) (target) determined by FIM method
    c : ndarray of float [n_ele], optional, default: None
        Congruence factor map normalized to 1 (whole ROI) used to weight the difference between the optimal e-field
        and the candidate e-field. If None, no weighting is applied.

    Returns
    -------
    res : ndarray of float [n_combs]
        Difference between e-fields and optimal e-field.
    """

    if c is None:
        c = np.ones(array.shape[1])

    res = np.zeros(len(idx_list))

    array = array[:, ele_idx_1]

    if array.ndim == 1:
        array = array[:, np.newaxis]

    for j, ind in enumerate(idx_list):
        res[j] = np.linalg.norm((e_opt - array[ind[-1], ele_idx_1]) * c[ele_idx_1])

    return res


def workhorse_fim_svd(idx_list, array, ele_idx_1, ele_idx_2, e_opt, c=None):
    """
    Determine difference between e-fields and optimal e-field determined using the Fisher Information Matrix and
    condition number.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    ele_idx_1 : ndarray of float [n_roi_1]
        Element indices for which the fim optimization is performed for
    ele_idx_2 : ndarray of float [n_roi_2]
        Element indices for which the svd optimization is performed for
    e_opt : float
        Optimal electric field value (target) determined by FIM method
    c : ndarray of float [n_ele], optional, default: None
        Congruence factor map normalized to 1 (whole ROI) used to weight the difference between the optimal e-field
        and the candidate e-field. If None, no weighting is applied.

    Returns
    -------
    res_fim : ndarray of float [n_combs]
        Difference between e-fields and optimal e-field.
    res_svd : ndarray of float [n_combs]
        Condition number. Lower values indicate more orthogonal e-field combinations (better)
    """

    if c is None:
        c = np.ones(array.shape[1])

    res_fim = np.zeros(len(idx_list))
    res_svd = np.zeros(len(idx_list))

    array_fim = array[:, ele_idx_1]

    if array_fim.ndim == 1:
        array_fim = array_fim[:, np.newaxis]

    array_svd = array[:, ele_idx_2]

    if array_svd.ndim == 1:
        array_svd = array_svd[:, np.newaxis]

    for j, ind in enumerate(idx_list):
        # fim
        intensity = np.mean(array_fim[ind[-1], :] / e_opt)
        res_fim[j] = np.linalg.norm((e_opt - intensity * array_fim[ind[-1], :]) * c[ele_idx_1])

        # svd
        u, s, vh = scipy.linalg.svd(array_svd[ind, :])
        res_svd[j] = np.max(s)/np.min(s)

    return res_fim, res_svd


def workhorse_fim_mc(idx_list, array, ele_idx_1, ele_idx_2, e_opt, c=None, mode="rows"):
    """
    Determine difference between e-fields and optimal e-field determined using the Fisher Information Matrix and
    mutual coherence.

    Parameters
    ----------
    idx_list : list of lists [n_combs][n_zaps]
        Index lists of zaps containing different possible combinations. Usually only the last index changes.
    array : ndarray of float [n_zaps x n_ele]
        Electric field for different coil positions and elements
    ele_idx_1 : ndarray of float [n_roi_1]
        Element indices for which the fim optimization is performed for
    ele_idx_2 : ndarray of float [n_roi_2]
        Element indices for which the mc optimization is performed for
    e_opt : float
        Optimal electric field value (target) determined by FIM method
    c : ndarray of float [n_ele], optional, default: None
        Congruence factor map normalized to 1 (whole ROI) used to weight the difference between the optimal e-field
        and the candidate e-field. If None, no weighting is applied.

    Returns
    -------
    res_fim : ndarray of float [n_combs]
        Difference between e-fields and optimal e-field.
    res_mc : ndarray of float [n_combs]
        Mutual coherence. Lower values indicate more orthogonal e-field combinations (better)
    """

    if c is None:
        c = np.ones(array.shape[1])

    res_fim = np.zeros(len(idx_list))
    res_mc = np.zeros(len(idx_list))

    array_fim = array[:, ele_idx_1]

    if array_fim.ndim == 1:
        array_fim = array_fim[:, np.newaxis]

    array_mc = array[:, ele_idx_2]

    if array_mc.ndim == 1:
        array_mc = array_mc[:, np.newaxis]

    for j, ind in enumerate(idx_list):
        # fim
        res_fim[j] = np.linalg.norm((e_opt - array_fim[ind[-1], :]) * c[ele_idx_1])

        # mc
        if mode == "cols":
            res_mc[j] = pynibs.mutual_coherence(array_mc[ind, :])
        elif mode == "rows":
            res_mc[j] = pynibs.mutual_coherence(array_mc[ind, :].transpose())
        else:
            raise NotImplementedError("Specified mode not implemented. Choose 'rows' or 'cols'.")

    return res_fim, res_mc


def get_optimal_coil_positions(e_matrix, criterion, n_stim, ele_idx_1=None, ele_idx_2=None, fn_out_hdf5=None, n_cpu=4,
                               zap_idx_opt=None, fun=None, p=None, c=None, weights=[0.5, 0.5],
                               overwrite=True, verbose=True, fn_coilpos_hdf5=None, start_zap_idx=0):
    """
    Determine set of optimal coil positions for TMS regression analysis.

    Parameters
    ----------
    e_matrix : ndarray of float [n_stim, n_ele]
        Matrix containing the electric field values in the ROI
    criterion : str
        Optimization criterion:
        - "mc_cols": Minimization of mutual coherence between columns
        - "mc_rows": Minimization of mutual coherence between rows
        - "svd": Minimization of condition number
        - "dist": Equal distant sampling
        - "dist_svd": Minimization of condition number and equidistant sampling
        - "dist_mc_cols": Minimization of mutual coherence between columns and equidistant sampling
        - "dist_mc_rows": Minimization of mutual coherence between rows and equidistant sampling
        - "coverage": Maximizes the electric field coverage
        - "variability": Maximizes variability between elements
    n_stim : int
        Maximum number of stimulations
    ele_idx_1 : ndarray of int, optional, default: None
        Element indices the first optimization goal is performed for, If None, all elements are consiered
    ele_idx_2 : ndarray of int, optional, default: None
        Element indices the first optimization goal is performed for. If None, all elements are consiered
    n_cpu : int
        Number of threads
    fn_out_hdf5 : str, optional, default: None
        Returns the list of optimal zap indices if fn_out_hdf5 is None, otherwise, save the results in .hdf5 file.
        Filename of output .hdf5 file where the zap index lists are saved in subfolder "zap_index_lists"
        - "zap_index_lists/0": [213]
        - "zap_index_lists/1": [213, 5]
        - etc
    zap_idx_opt : list of int, optional, default: None
        List of already selected optimal coil positions
        (those are ignored in the optimization and will not be picked again)
    fun : function object
        Function object defined in interval [0, 1]. (only needed for fim optimization)
    p : list of dict [n_ele], optional, default: None
        List of dicts containing the parameter estimates (whole ROI). The keys are the parameter names of fun.
        (only needed for fim and dist optimization)
    c : ndarray of float [n_ele], optional, default: None
        Congruence factor in each ROI element. Used to weight fim and dist optimization.
        (only needed for fim and dist optimization)
    weights : list of float [2], optional, default: [0.5, 0.5]
        Weights of optimization criteria in case of multiple goal functions (e.g. fim_svd). Higher weight means higher
        importance for the respective criteria. By default both optimization criteria are weighted equally [0.5, 0.5].
    overwrite : bool, optional, default: True
        Overwrite existing solutions or read existing hdf5 file and continue optimization
    verbose : bool, optional, default: True
        Print output messages
    fn_coilpos_hdf5 : str
        File containing the corresponding coil positions and orientations (centers, m0, m1, m2)
    start_zap_idx : int, optional, default: 0
        First zap index to start greedy search

    Returns
    -------
    zap_idx_e_opt : list of int [n_stim]
        Optimal zap indices
    <File> .hdf5 file
        Output file containing the zap index lists
    """

    if zap_idx_opt is not None and fn_out_hdf5 is not None and (os.path.exists(fn_out_hdf5) and not overwrite):
        raise ValueError("zap_idx_opt and fn_out_hdf5 given... please choose whether to load optimal zap indices from "
                         "file or given explicitly as list")

    e = e_matrix
    x = np.zeros((1, 1))
    y = np.zeros((1, 1))

    if ele_idx_1 is None:
        ele_idx_1 = np.arange(e.shape[1])

    if ele_idx_2 is None:
        ele_idx_2 = np.arange(e.shape[1])

    if c is not None:
        c = c / np.max(c)

    zap_idx_e_opt = [0 for _ in range(n_stim)]
    crit = np.zeros(n_stim)

    if zap_idx_opt is not None:
        kmax = len(zap_idx_opt)-1
        zap_idx_e_opt[len(zap_idx_opt)-1] = zap_idx_opt
        idx_rem = [i for i in range(e.shape[0]) if i not in zap_idx_e_opt[int(kmax)]]
    else:
        kmax = 0
        zap_idx_e_opt[0] = [start_zap_idx]
        idx_rem = [i for i in range(1, e.shape[0])]

    if fn_coilpos_hdf5 is not None:
        with h5py.File(fn_coilpos_hdf5, "r") as f:
            centers = f["centers"][:]
            m0 = f["m0"][:]
            m1 = f["m1"][:]
            m2 = f["m2"][:]

    pool = multiprocessing.Pool(n_cpu)

    if fn_out_hdf5 is not None:
        if not overwrite and os.path.exists(fn_out_hdf5):
            # load already present optimization results
            with h5py.File(fn_out_hdf5, "r") as f:

                kmax = 0

                try:
                    keys = f[f"zap_index_lists"].keys()
                    crit = f["criterion"][:]

                    for k in keys:
                        zap_idx_e_opt[int(k)] = list(f[f"zap_index_lists/{k}"][:])

                        if int(k) > kmax:
                            kmax = int(k)

                    # check if loaded sequence starts with the same zap index as intended
                    if zap_idx_e_opt[0] != [start_zap_idx]:
                        kmax = 0
                        zap_idx_e_opt = [0 for _ in range(n_stim)]
                        zap_idx_e_opt[0] = [start_zap_idx]
                        if verbose:
                            print(f"Loaded sequence does not start with specified start idx (restarting optimization)")
                            print(f"=================================================================================")
                    else:
                        if verbose:
                            print(f"Loading optimal index set for n={kmax+1}")
                            print(f"====================================")

                except KeyError:
                    pass

                idx_rem = [i for i in range(e.shape[0]) if i not in zap_idx_e_opt[kmax]]

        else:
            # save first zap
            with h5py.File(fn_out_hdf5, "w") as f:
                f.create_dataset(f"zap_index_lists/0", data=zap_idx_e_opt[0])

    for i in range(kmax+1, n_stim):
        if verbose:
            # if kmax == 0:
            #     print(f"Initializing greedy algorithm for n={i}")
            #     print(f"=====================================")
            #     print(f" >>> Chosen index: {start_zap_idx}")

            print(f"Calculating optimal idx for n={i+1}")
            print(f"==================================")

        # prepare calculations for "coverage" criterion
        if criterion == "coverage":
            workhorse_prepare = partial(workhorse_coverage_prepare, array=e, zap_idx=zap_idx_e_opt[i-1])
            ele_idx_list_chunks = pynibs.compute_chunks([j for j in range(e.shape[1])], n_cpu)
            res = pool.map(workhorse_prepare, ele_idx_list_chunks)

            for j in range(len(res)):
                if j == 0:
                    x = res[0][0]
                    y = res[0][1]
                else:
                    x = np.hstack((x, res[j][0]))
                    y = np.hstack((y, res[j][1]))

            workhorse_partial = partial(workhorse_coverage, array=e, x=x, y=y)

        if "fim" in criterion:

            # normalize e-field to [0, 1]
            x = (e[zap_idx_opt, :][:, ele_idx_1] - np.min(e[:, ele_idx_1], axis=0)) / \
                (np.max(e[:, ele_idx_1], axis=0) - np.min(e[:, ele_idx_1], axis=0))

            # determine optimal e-fields for next zap
            e_opt_norm = np.zeros(len(ele_idx_1))
            for i_ele, e_idx_1 in enumerate(ele_idx_1):
                e_opt_norm[i_ele] = pynibs.get_optimal_sample_fim(fun=fun, x=x[:, i_ele], p=p[e_idx_1])

            e_opt = e_opt_norm * (np.max(e[:, ele_idx_1], axis=0) - np.min(e[:, ele_idx_1], axis=0)) + \
                    np.min(e[:, ele_idx_1], axis=0)

        if criterion == "svd":
            workhorse_partial = partial(workhorse_svd, array=e, ele_idx_1=ele_idx_1)

        elif criterion == "dist":
            workhorse_partial = partial(workhorse_dist, array=e, ele_idx_1=ele_idx_1)

        elif criterion == "mc_cols":
            e = e - np.mean(e[:, ele_idx_1], axis=1)[:, np.newaxis]
            workhorse_partial = partial(workhorse_mc, array=e, ele_idx_1=ele_idx_1, mode="cols")

        elif criterion == "mc_rows":
            workhorse_partial = partial(workhorse_mc, array=e, ele_idx_1=ele_idx_1, mode="rows")

        elif criterion == "variability":
            workhorse_partial = partial(workhorse_variability, array=e, ele_idx_1=ele_idx_1)

        elif criterion == "dist_mc_cols":
            if weights[0] == 0:
                workhorse_partial = partial(workhorse_mc, array=e, ele_idx_1=ele_idx_2, mode="cols")
            elif weights[1] == 0:
                workhorse_partial = partial(workhorse_dist, array=e, ele_idx_1=ele_idx_1)
            else:
                workhorse_partial = partial(workhorse_dist_mc, array=e, ele_idx_1=ele_idx_1, ele_idx_2=ele_idx_2, mode="cols")

        elif criterion == "dist_mc_rows":
            if weights[0] == 0:
                workhorse_partial = partial(workhorse_mc, array=e, ele_idx_1=ele_idx_2, mode="rows")
            elif weights[1] == 0:
                workhorse_partial = partial(workhorse_dist, array=e, ele_idx_1=ele_idx_1)
            else:
                workhorse_partial = partial(workhorse_dist_mc, array=e, ele_idx_1=ele_idx_1, ele_idx_2=ele_idx_2, mode="rows")

        elif criterion == "dist_svd":
            if weights[0] == 0:
                workhorse_partial = partial(workhorse_svd, array=e, ele_idx_1=ele_idx_2)
            elif weights[1] == 0:
                workhorse_partial = partial(workhorse_dist, array=e, ele_idx_1=ele_idx_1)
            else:
                workhorse_partial = partial(workhorse_dist_svd, array=e, ele_idx_1=ele_idx_1, ele_idx_2=ele_idx_2)

        elif criterion == "fim":
            workhorse_partial = partial(workhorse_fim, array=e, ele_idx_1=ele_idx_1, e_opt=e_opt, c=c)

        elif criterion == "fim_svd":
            if weights[0] == 0:
                workhorse_partial = partial(workhorse_svd, array=e, ele_idx_1=ele_idx_2)
            elif weights[1] == 0:
                workhorse_partial = partial(workhorse_fim, array=e, ele_idx_1=ele_idx_1, e_opt=e_opt, c=c)
            else:
                workhorse_partial = partial(workhorse_fim_svd, array=e, ele_idx_1=ele_idx_1, ele_idx_2=ele_idx_2,
                                            e_opt=e_opt, c=c)

        elif criterion == "fim_mc_rows":
            if weights[0] == 0:
                workhorse_partial = partial(workhorse_mc, array=e, ele_idx_1=ele_idx_2, mode="rows")
            elif weights[1] == 0:
                workhorse_partial = partial(workhorse_fim, array=e, ele_idx_1=ele_idx_1, e_opt=e_opt, c=c)
            else:
                workhorse_partial = partial(workhorse_fim_mc, array=e, ele_idx_1=ele_idx_1, ele_idx_2=ele_idx_2,
                                            e_opt=e_opt, c=c, mode="rows")

        elif criterion == "fim_mc_cols":
            if weights[0] == 0:
                workhorse_partial = partial(workhorse_mc, array=e, ele_idx_1=ele_idx_2, mode="cols")
            elif weights[1] == 0:
                workhorse_partial = partial(workhorse_fim, array=e, ele_idx_1=ele_idx_1, e_opt=e_opt, c=c)
            else:
                workhorse_partial = partial(workhorse_fim_mc, array=e, ele_idx_1=ele_idx_1, ele_idx_2=ele_idx_2,
                                            e_opt=e_opt, c=c, mode="cols")

        else:
            raise NameError(f"criterion: {criterion} not implemented")

        # determine criterion
        idx_list = []
        for j in range(len(idx_rem)):
            idx_list.append(zap_idx_e_opt[i-1] + [idx_rem[j]])

        idx_list_chunks = pynibs.compute_chunks(idx_list, n_cpu)
        res = pool.map(workhorse_partial, idx_list_chunks)

        # extract best solution (multiple objectives)
        if type(res[0]) is tuple:

            for j in range(len(res)):
                if j == 0:
                    res_all = np.vstack(res[j]).transpose()
                else:
                    res_all = np.vstack((res_all, np.vstack(res[j]).transpose()))

            # filter nans # TODO: not to 1e6 -> max of each opt criterion (column)
            res_all[np.isnan(res_all)] = 1e6
            res_all[res_all == 0] = 1e-6

            # normalize both optimization criteria to [0, 1]
            res_all = (res_all - np.min(res_all, axis=0)) / (np.max(res_all, axis=0) - np.min(res_all, axis=0))

            # weight optimization criteria
            res_all = res_all * weights

            # find best solution with lowest sum
            res_all_sum = np.sum(res_all, axis=1)
            idx_best = np.argmin(res_all_sum)
            crit[i] = res_all_sum[idx_best]

        # extract best solution (single objective)
        else:
            # filter nans
            res = np.concatenate(res)
            res[np.isnan(res)] = 1e6

            # find best solution
            idx_best = np.argmin(res)
            crit[i] = res[idx_best]

        if verbose:
            print(f" >>> Best index: {idx_rem[idx_best]}, criterion: {crit[i]}")

        zap_idx_e_opt[i] = zap_idx_e_opt[i-1] + [idx_rem[idx_best]]
        idx_rem = [k for k in idx_rem if k != idx_rem[idx_best]]

        if fn_out_hdf5 is not None:
            # save results
            with h5py.File(fn_out_hdf5, "a") as f:
                try:
                    del f["criterion"]
                except (RuntimeError, KeyError):
                    f.create_dataset("criterion", data=crit)

                try:
                    f.create_dataset(f"zap_index_lists/{i}", data=zap_idx_e_opt[i])
                except (RuntimeError, KeyError):
                    if overwrite:
                        del f[f"zap_index_lists/{i}"]
                        f.create_dataset(f"zap_index_lists/{i}", data=zap_idx_e_opt[i])
                    else:
                        print(f"Could not write zap_index_lists/{i}. Dataset already exists.")

                if fn_coilpos_hdf5 is not None:
                    try:
                        del f["centers"], f["m0"], f["m1"], f["m2"]
                    except (RuntimeError, KeyError):
                        if overwrite:
                            f.create_dataset("centers", data=centers[zap_idx_e_opt[i], :])
                            f.create_dataset("m0", data=m0[zap_idx_e_opt[i], :])
                            f.create_dataset("m1", data=m1[zap_idx_e_opt[i], :])
                            f.create_dataset("m2", data=m2[zap_idx_e_opt[i], :])

    pool.close()
    pool.join()

    if fn_out_hdf5 is None:
        return zap_idx_e_opt[-1]
    else:
        with h5py.File(fn_out_hdf5, "a") as f:
            try:
                f.create_dataset(f"zap_index_list", data=np.array(zap_idx_e_opt[-1])[:, np.newaxis])
            except (RuntimeError, KeyError):
                if overwrite:
                    del f[f"zap_index_list"]
                    f.create_dataset(f"zap_index_list", data=np.array(zap_idx_e_opt[-1])[:, np.newaxis])
                else:
                    print(f"Could not write zap_index_list. Dataset already exists.")

            if fn_coilpos_hdf5 is not None:
                m0_opt_reshaped = np.hstack((m0[zap_idx_e_opt[-1], :],
                                             np.zeros((len(zap_idx_e_opt[-1]), 1)))).T[:, np.newaxis, :]
                m1_opt_reshaped = np.hstack((m1[zap_idx_e_opt[-1], :],
                                             np.zeros((len(zap_idx_e_opt[-1]), 1)))).T[:, np.newaxis, :]
                m2_opt_reshaped = np.hstack((m2[zap_idx_e_opt[-1], :],
                                             np.zeros((len(zap_idx_e_opt[-1]), 1)))).T[:, np.newaxis, :]
                centers_opt_reshaped = np.hstack((centers[zap_idx_e_opt[-1], :],
                                                  np.ones((len(zap_idx_e_opt[-1]), 1)))).T[:, np.newaxis, :]
                matsimnibs = np.concatenate((m0_opt_reshaped,
                                             m1_opt_reshaped,
                                             m2_opt_reshaped,
                                             centers_opt_reshaped), axis=1)

                try:
                    f.create_dataset("centers", data=centers[zap_idx_e_opt[-1], :])
                    f.create_dataset("m0", data=m0[zap_idx_e_opt[-1], :])
                    f.create_dataset("m1", data=m1[zap_idx_e_opt[-1], :])
                    f.create_dataset("m2", data=m2[zap_idx_e_opt[-1], :])
                    f.create_dataset("matsimnibs", data=matsimnibs)
                except (RuntimeError, KeyError):
                    if overwrite:
                        del f["centers"], f["m0"], f["m1"], f["m2"]
                        f.create_dataset("centers", data=centers[zap_idx_e_opt[-1], :])
                        f.create_dataset("m0", data=m0[zap_idx_e_opt[-1], :])
                        f.create_dataset("m1", data=m1[zap_idx_e_opt[-1], :])
                        f.create_dataset("m2", data=m2[zap_idx_e_opt[-1], :])
                        f.create_dataset("matsimnibs", data=matsimnibs)
                    else:
                        print(f"Could not write centers, m0, m1, m2 ... Dataset already exists.")


def online_optimization(fn_subject_hdf5, fn_roi_ss_indices_hdf5, fn_out_hdf5, fn_stimsites_hdf5, e_matrix, mep,
                        mesh_idx, roi_idx, n_zaps_init=3, criterion_init="mc_rows", criterion="coverage", n_cpu=4,
                        threshold=0.8, weights=[0.5, 0.5], eps0=0.01, eps0_dist=1, exponent=5, perc=99,
                        n_refit=0, fun=sigmoid, verbose=True):
    """
    Performs virtual online optimization to determine the congruence factor. After an initial set of coil positions,
    the algorithm iteratively optimizes the next coil position based on the virtually measured MEP data.

    Parameters
    ----------
    fn_subject_hdf5 : str
        Filename of subject .hdf5 file
    fn_roi_ss_indices_hdf5 : str
        Filename of .hdf5 file containing the element indices of the subsampled ROI in f["roi_indices"]
    e_matrix : ndarray of float [n_zaps x n_ele]
        Electric field matrix
    mep : ndarray of float [n_zaps]
        Motor evoked potentials for every stimulation
    fn_out_hdf5 : str
        Filename of .hdf5 output file containing the coil positions and the congruence factor maps for every iteration
    fn_stimsites_hdf5 : str
        Filename of the .hdf5 file containing the stimulation sites in "centers", "m0", "m1", "m2"
    mesh_idx : int
        Mesh index
    roi_idx : int
        ROI index
    n_zaps_init : int, optional, default: 3
        Number of initial samples optimized using optimization criterion specified in "criterion_init"
    criterion_init : str, optional, default: "mc_rows"
        Optimization criterion for which the initial samples are optimized (e.g. "mc_rows", "svd", ...)
    criterion : str, optional, default: "coverage"
        Optimization criterion for which the online optimization is performed (e.g. "coverage", "mc_rows", "svd", ...)
    n_cpu : int, optional, dfault: 4
        Number of CPU cores to use
    threshold : float, optional, default: 0.1
        Threshold between [0 ... 1] of the maximal congruence factor. Elements where c > threshold * max(c)
        are included in the online optimization to select the next optimal coil position.
    weights : list of float [2], optional, default: [0.5, 0.5]
        Weights of optimization criteria in case of multiple goal functions (e.g. fim_svd). Higher weight means higher
        importance for the respective criteria. By default both optimization criteria are weighted equally [0.5, 0.5].
    eps0 : float, optional, default: 0.01
        First error threshold to terminate the online optimization. The normalized root mean square deviation is
        calculated between the current and the previous solution. If the error is lower than eps0 for 3 times in a row,
        the online optimization terminates and returns the results.
    eps0_dist : float, optional, default: 1
        Second error threshold to terminate the online optimization. The geodesic distance in mm of the hotspot is
        calculated between the current and the previous solution. If the error is lower than eps0_dist for 3 times
        in a row, the online optimization terminates and returns the results.
    exponent : float, optional, default: 5
        Exponent the congruence factor map is scaled c**exponent
    perc : float, optional, default: 99
        Percentile the congruence factor map is normalized (between 0 and 100)
    n_refit : int, optional, default: 0
        Number of refit iterations. No refit is applied if n_refit=0.
    fun : function object, optional, default: pynibs.linear
        Function to use to determine the congruence factor (e.g. pynibs.linear, pynibs.sigmoid, ...)
    verbose : bool, optional, default: True
        Plot output messages

    Returns
    -------
    <file> .hdf5 file
        Results output file containing the coil positions and the congruence factor maps for every iteration

    """
    print("Starting online congruence factor optimization:")
    print("===============================================")
    print(f" > fn_subject_hdf5:          {fn_subject_hdf5}")
    print(f" > fn_roi_ss_indices_hdf5:   {fn_roi_ss_indices_hdf5}")
    print(f" > fn_stimsites_hdf5:        {fn_stimsites_hdf5}")
    print(f" > fn_out_hdf5:              {fn_out_hdf5}")
    print(f" > e_matrix:                 shape: {e_matrix.shape}")
    print(f" > mep:                      shape: {mep.shape}")
    print(f" > mesh_idx:                 {mesh_idx}")
    print(f" > roi_idx:                  {roi_idx}")
    print(f" > n_zaps_init:              {n_zaps_init}")
    print(f" > criterion_init:           {criterion_init}")
    print(f" > criterion:                {criterion}")
    print(f" > n_cpu:                    {n_cpu}")
    print(f" > threshold:                {threshold}")
    print(f" > weights:                  {weights}")
    print(f" > eps0:                     {eps0}")
    print(f" > eps0_dist:                {eps0_dist}")
    print(f" > exponent:                 {exponent}")
    print(f" > perc:                     {perc}")
    print(f" > n_refit:                  {n_refit}")
    print(f" > fun:                      {fun.__name__}")
    print(f" > verbose:                  {verbose}")
    print("")

    zap_idx = dict()
    c = dict()

    # load subject
    if verbose:
        print(f"Loading subject")
    subject = pynibs.load_subject(fn_subject_hdf5)

    # load ROI and perform subsampling
    if verbose:
        print(f"Loading ROI and perform subsampling")
    roi = pynibs.load_roi_surface_obj_from_hdf5(subject.mesh[mesh_idx]['fn_mesh_hdf5'])[roi_idx]
    con = roi.node_number_list
    points = roi.node_coord_mid

    with h5py.File(fn_roi_ss_indices_hdf5, "r") as f:
        ele_idx_ss = f["roi_indices"][:]

    # e-fields
    if verbose:
        print(f"Loading electric field from regression.hdf5")
    n_ele = e_matrix.shape[1]

    # loading coil positions and create matsimnibs [4x4] matrices
    with h5py.File(fn_stimsites_hdf5, "r") as f:
        centers_all = f["centers"][:]
        m0_all = f["m0"][:]
        m1_all = f["m1"][:]
        m2_all = f["m2"][:]

    coil_mean = dict()
    current_dict = dict()

    for i in range(centers_all.shape[0]):
        coil_mean[str(i)] = np.hstack((m0_all[i, :][:, np.newaxis],
                                       m1_all[i, :][:, np.newaxis],
                                       m2_all[i, :][:, np.newaxis],
                                       centers_all[i, :][:, np.newaxis]))
        current_dict[str(i)] = 1

    # determine initial number of optimal samples
    if verbose:
        print(f"Determine optimal coil positions for initial number of {n_zaps_init} samples using {criterion_init}")

    zap_idx_opt = pynibs.get_optimal_coil_positions(e_matrix=e_matrix,
                                                    ele_idx_1=ele_idx_ss,
                                                    ele_idx_2=None,
                                                    criterion=criterion_init,
                                                    n_stim=n_zaps_init,
                                                    fn_out_hdf5=None,
                                                    n_cpu=n_cpu,
                                                    zap_idx_opt=None,
                                                    weights=weights,
                                                    overwrite=False,
                                                    verbose=True)

    # determine initial c-factor map for all N (not existing in real life)
    if verbose:
        print(f"Determine reference c-factor map (N)")

    c_ref_N = regress_data(elm_idx_list=np.arange(n_ele),
                           e_matrix=e_matrix,
                           mep=mep,
                           zap_idx=None,
                           fun=fun,
                           n_refit=n_refit,
                           n_cpu=n_cpu,
                           con=con,
                           return_fits=False,
                           refit_discontinuities=True)

    ref_N = c_ref_N.flatten()**exponent
    ref_N = ref_N / np.percentile(ref_N, perc)

    c_max_idx_N = np.argmax(c_ref_N)

    ##########################################################################
    #
    # Robot measures initial offline optimal coil positions and collects MEPs
    #
    ##########################################################################

    # determine initial c-factor map (after robot measured first offline optimal coil positions)
    if verbose:
        print(f"Determine initial c-factor map")

    c_init, p = regress_data(elm_idx_list=np.arange(n_ele),
                             e_matrix=e_matrix,
                             mep=mep,
                             zap_idx=zap_idx_opt,
                             fun=fun,
                             n_refit=n_refit,
                             n_cpu=n_cpu,
                             con=con,
                             return_fits=True,
                             refit_discontinuities=True)
    ref = c_init.flatten()**exponent
    ref = ref / np.percentile(ref, perc)

    eps = [eps0 + 1 for _ in range(len(zap_idx_opt))]
    eps_N = [eps0 + 1 for _ in range(len(zap_idx_opt))]

    gdist = [eps0_dist + 1 for _ in range(len(zap_idx_opt))]
    gdist_N = [eps0_dist + 1 for _ in range(len(zap_idx_opt))]

    for i in range(len(zap_idx_opt)):
        zap_idx[str(i)] = zap_idx_opt[:(i+1)]
        c[str(i)] = np.zeros(ref.shape)

    c[str(n_zaps_init-1)] = c_init

    n_zaps = [i for i in range(1, n_zaps_init+1)]

    # Start online optimization loop
    while not ((np.array(eps[-3:]) < eps0).all() and (np.array(gdist[-3:]) < eps0_dist).all()):

        if "fim" in criterion or "dist_" in criterion:
            # find elements with values greater than threshold
            mask_perc = ref >= threshold*np.max(ref)
            ele_idx_1 = np.where(mask_perc)[0]
            ele_idx_2 = ele_idx_ss

        else:
            ele_idx_1 = ele_idx_ss
            ele_idx_2 = ele_idx_ss

        # optimize coil positions for subset of ROI elements
        if verbose:
            print(f"Optimizing next coil position for ROI_1: {len(ele_idx_1)} / ROI_2: {len(ele_idx_2)} elements "
                  f"using {criterion}")

        n_zaps.append(n_zaps[-1] + 1)
        zap_idx_opt = pynibs.get_optimal_coil_positions(e_matrix=e_matrix,
                                                        ele_idx_1=ele_idx_1,
                                                        ele_idx_2=ele_idx_2,
                                                        criterion=criterion,
                                                        n_stim=n_zaps[-1],
                                                        fn_out_hdf5=None,
                                                        n_cpu=n_cpu,
                                                        zap_idx_opt=zap_idx_opt,
                                                        fun=fun,
                                                        p=p,
                                                        c=ref,
                                                        weights=weights,
                                                        overwrite=False,
                                                        verbose=True)
        key = str(len(zap_idx_opt)-1)
        zap_idx[key] = zap_idx_opt

        ##########################################################################
        #
        # Robot measures next optimal coil position
        #
        ##########################################################################

        # determine updated c-factor map
        if verbose:
            print(f"Determine c-factor map for {len(zap_idx_opt)} zaps")

        c[key], p = regress_data(elm_idx_list=np.arange(n_ele),
                                 e_matrix=e_matrix,
                                 mep=mep,
                                 zap_idx=zap_idx_opt,
                                 fun=fun,
                                 n_refit=n_refit,
                                 n_cpu=n_cpu,
                                 con=con,
                                 return_fits=True,
                                 refit_discontinuities=True)
        arr = c[key].flatten()**exponent
        arr = arr / np.percentile(arr, perc)

        ##########################################################################
        #
        # Plot updated c-factor map
        #
        ##########################################################################

        # determine NRMSD w.r.t. previous solution
        eps.append(pynibs.nrmsd(arr, ref))

        if verbose:
            print(f"NRMSD to previous solution: {eps[-1]}")

        # determine NRMSD w.r.t. global solution (not existing in real life)
        eps_N.append(pynibs.nrmsd(arr, ref_N))

        if verbose:
            print(f"NRMSD to global solution: {eps_N[-1]}")

        # determine geodesic distance w.r.t. previous solution
        nodes_dist, tris_dist = pynibs.geodesic_dist(nodes=points, tris=con, source=np.argmax(c[key]), source_is_node=False)
        gdist.append(tris_dist[np.argmax(c[str(len(zap_idx_opt)-2)])])

        if verbose:
            print(f"GDIST to previous solution: {gdist[-1]:.3f} mm")

        # determine geodesic distance w.r.t. global solution (not existing in real life)
        gdist_N.append(tris_dist[c_max_idx_N])

        if verbose:
            print(f"GDIST to global solution: {gdist_N[-1]:.3f} mm")

        # set current solution as ref
        ref = arr

    eps_N = np.array(eps_N)
    eps = np.array(eps)
    gdist_N = np.array(gdist_N)
    gdist = np.array(gdist)
    n_zaps = np.array(n_zaps)

    if verbose:
        print(f"Saving results to {fn_out_hdf5}")

    with h5py.File(fn_out_hdf5, "w") as f:
        f.create_dataset("nrmsd", data=eps_N)
        f.create_dataset("nrmsd_n_1", data=eps)
        f.create_dataset("gdist", data=gdist_N)
        f.create_dataset("gdist_n_1", data=gdist)
        f.create_dataset("n_zaps", data=n_zaps)
        f.create_dataset(f"c_ref", data=c_ref_N)

        for key in zap_idx:
            f.create_dataset(f"zap_index_lists/{key}", data=zap_idx[key])
            f.create_dataset(f"c/{key}", data=c[key])

            n_zaps_tmp = len(zap_idx[key])
            centers = np.zeros((n_zaps_tmp, 3))
            m0 = np.zeros((n_zaps_tmp, 3))
            m1 = np.zeros((n_zaps_tmp, 3))
            m2 = np.zeros((n_zaps_tmp, 3))
            current = np.zeros((n_zaps_tmp, 1))

            for i, j in enumerate(zap_idx[key]):
                centers[i, :] = coil_mean[str(j)][0:3, 3]
                m0[i, :] = coil_mean[str(j)][0:3, 0]
                m1[i, :] = coil_mean[str(j)][0:3, 1]
                m2[i, :] = coil_mean[str(j)][0:3, 2]
                current[i, 0] = current_dict[str(j)]

            f.create_dataset(f"centers/{key}", data=centers)
            f.create_dataset(f"m0/{key}", data=m0)
            f.create_dataset(f"m1/{key}", data=m1)
            f.create_dataset(f"m2/{key}", data=m2)
            f.create_dataset(f"current/{key}", data=current)

    # create geo .hdf5
    fn_geo_hdf5 = os.path.splitext(fn_out_hdf5)[0] + "_geo.hdf5"
    pynibs.write_geo_hdf5_surf(out_fn=fn_geo_hdf5,
                               points=points,
                               con=con,
                               replace=True,
                               hdf5_path='/mesh')

    # write xdmf file with optimal results
    if verbose:
        print(f"Creating .xdmf ...")

    pynibs.write_temporal_xdmf(hdf5_fn=fn_out_hdf5,
                               data_folder='c',
                               coil_center_folder="centers",
                               coil_ori_0_folder="m0",
                               coil_ori_1_folder="m1",
                               coil_ori_2_folder="m2",
                               coil_current_folder="current",
                               hdf5_geo_fn=fn_geo_hdf5,
                               overwrite_xdmf=True,
                               verbose=False)

    # plot results
    fn_plot_nrmsd = os.path.splitext(fn_out_hdf5)[0] + "_nrmsd.png"
    fn_plot_gdist = os.path.splitext(fn_out_hdf5)[0] + "_gdist.png"

    if verbose:
        print(f"Plotting results to {fn_plot_nrmsd}")

    sort_idx = np.argsort(n_zaps)
    n_zaps = n_zaps[sort_idx]
    eps_N = eps_N[sort_idx]
    eps = eps[sort_idx]

    # nrmsd (n vs N) error
    plt.plot(n_zaps[n_zaps_init:], eps_N[n_zaps_init:], color="r")

    # nrmsd (n vs n-1) error
    plt.plot(n_zaps[n_zaps_init:], eps[n_zaps_init:], color="b")

    # 5% error bar
    plt.plot(np.array([n_zaps[4], n_zaps[-1]]), np.array([0.05, 0.05]), "r--")

    # 1% error bar
    plt.plot(np.array([n_zaps[4], n_zaps[-1]]), np.array([0.01, 0.01]), "b--")

    plt.grid()
    plt.xlabel("n", fontsize=11)
    plt.ylabel("NRMSD", fontsize=11)
    plt.ylim([0, 0.2])
    # plt.xscale("log")
    plt.title("Convergence analysis of online optimization", fontsize=11)
    plt.legend(["n vs N", "n vs (n-1)"])
    plt.savefig(fn_plot_nrmsd)
    plt.close()

    # gdist (n vs N) error
    plt.plot(n_zaps[n_zaps_init:], gdist_N[n_zaps_init:], color="r")

    # gdist (n vs n-1) error
    plt.plot(n_zaps[n_zaps_init:], gdist[n_zaps_init:], color="b")

    # 1% error bar
    plt.plot(np.array([n_zaps[4], n_zaps[-1]]), np.array([1, 1]), "b--")

    plt.grid()
    plt.xlabel("n", fontsize=11)
    plt.ylabel("GDIST", fontsize=11)
    plt.title("Convergence analysis of online optimization", fontsize=11)
    plt.legend(["n vs N", "n vs (n-1)"])
    plt.savefig(fn_plot_gdist)
    plt.close()

    if verbose:
        print("DONE")


def get_fim_sample(fun, x, p):
    """
    Get Fisher Information Matrix of one single sample.

    Parameters
    ----------
    fun : function object
        Function object the fisher information matrix is calculated for. The sample is passed as the first argument.
    x : float
        Sample passed to function
    p : dict
        Dictionary containing the parameter estimates. The keys are the parameter names of fun.

    Returns
    -------
    fim_matrix : ndarray of float [n_params x n_params]
        Fisher information matrix
    """

    # read function arguments
    params = p.keys()

    # determine gradient of function w.r.t. parameters using forward approximation
    dfdp = np.zeros(len(params))

    for i, para in enumerate(params):
        # copy original params dict
        p_dp = dict()

        for pa in params:
            p_dp[pa] = p[pa]

        # perturb parameter
        dp = p[para] / 1000
        p_dp[para] = p[para] + dp

        # determine gradient with forward approximation
        dfdp[i] = (fun(x, **p_dp) - fun(x, **p)) / dp

    fim_matrix = np.outer(dfdp, dfdp)

    return fim_matrix


def get_det_fim(x, fun, p, fim_matrix):
    """
    Updates the Fisher Information Matrix and returns the negative determinant based on the sample x.
    It is a score how much information the additional sample yields.

    Parameters
    ----------
    fun : function object
        Function object defined in interval [0, 1].
    x : float
        Single sample location (interval [0, 1])
    p : dict
        Dictionary containing the parameter estimates. The keys are the parameter names of fun.
    fim_matrix : ndarray of float [n_params x n_params]
        Fisher Information Matrix

    Returns
    -------
    det : float
        Determinant of the Fisher Information Matrix after adding sample x
    """
    fim_matrix_sample = fim_matrix + get_fim_sample(fun=fun, x=x, p=p)
    sign, logdet = np.linalg.slogdet(fim_matrix_sample)

    return -sign*np.exp(logdet)


def init_fim_matrix(fun, x, p):
    """
    Initializes the Fisher Information Matrix based on the samples given in x.

    Parameters
    ----------
    fun : function object
        Function object defined in interval [0, 1].
    x : ndarray of float
        Initial sample locations (interval [0, 1])
    p : dict
        Dictionary containing the parameter estimates. The keys are the parameter names of fun.

    Returns
    -------
    fim_matrix : ndarray of float [n_params x n_params]
        Fisher Information Matrix
    """
    fim_matrix = np.zeros((len(p), len(p)))

    for i in range(len(x)):
        fim_matrix += get_fim_sample(fun=fun, x=x[i], p=p)

    return fim_matrix


def get_optimal_sample_fim(fun, p, x=None):
    """
    Determines optimal location of next sample by maximizing the determinant of the Fisher Information Matrix.

    Parameters
    ----------
    fun : function object
        Function object (interval [0, 1]).
    x : ndarray of float, optional, default: None
        Previous sample locations (interval [0, 1]).
    p : dict
        Dictionary containing the parameter estimates. The keys are the parameter names of fun.

    Returns
    -------
    x_opt : float
        Optimal location of next sample (interval [0, 1]).
    """
    # initialize fim matrix with initial samples
    if x is None:
        fim_matrix = np.zeros((len(p), len(p)))
    else:
        fim_matrix = init_fim_matrix(fun=fun, x=x, p=p)

    # run optimization
    # res = minimize(fun=get_det_fim,
    #                method="trust-constr",
    #                bounds=((0, 1),),
    #                x0=np.array([0.5]),
    #                args=(fun, p, fim_matrix),
    #                options={"finite_diff_rel_step": 0.05, "xtol": 0.01})

    # res = minimize(fun=get_det_fim,
    #                 method="SLSQP",
    #                 x0=np.array([0.5]),
    #                 bounds=((0, 1),),
    #                 args=(fun, p, fim_matrix),
    #                 options={'disp': None, "eps": 0.001, "ftol": 1e-12}) #,
    # return res.x[0]

    x_bf = np.linspace(0, 1, 200)
    det = np.zeros(len(x_bf))
    for i in range(len(x_bf)):
        det[i] = get_det_fim(x=x_bf[i], fun=fun, p=p, fim_matrix=fim_matrix)

    x_opt = x_bf[np.argmin(det)]

    return x_opt
